import torch

from ptls.loss import PairwiseMarginRankingLoss


def test_best_loss():
    prediction = torch.tensor([0.4, 0.5, 0.6, 0.2, 0.3])
    label = torch.tensor([1, 1, 1, 0, 0])

    loss = PairwiseMarginRankingLoss()
    out = loss(prediction, label)
    assert 0. == out
    assert type(out) is torch.Tensor


def test_usual_loss():
    prediction = torch.tensor([0.2, 0.5, 0.6, 0.7, 0.3])
    label = torch.tensor([1, 1, 1, 0, 0])

    loss = PairwiseMarginRankingLoss()
    out = loss(prediction, label)
    assert 0. < out
    assert type(out) is torch.Tensor


def test_one_class():
    prediction = torch.tensor([0.2, 0.5, 0.6, 0.7, 0.3])
    label = torch.tensor([0, 0, 0, 0, 0])

    loss = PairwiseMarginRankingLoss()
    out = loss(prediction, label)

    assert 0. == out
    assert type(out) is torch.Tensor


def test_minimal_example():
    prediction = torch.tensor([0.2, 0.5])
    label = torch.tensor([0, 1])

    loss = PairwiseMarginRankingLoss()
    out = loss(prediction, label)

    assert 0. == out
    assert type(out) is torch.Tensor



from functools import partial
from pathlib import Path

import pandas as pd
import pytorch_lightning as pl
import torch
from sklearn.model_selection import train_test_split

from ptls.data_load.data_module.emb_data_module import train_data_loader
from ptls.data_load.datasets import MemoryMapDataset
from ptls.data_load.datasets.dataloaders import inference_data_loader
from ptls.data_load.iterable_processing import ISeqLenLimit, SeqLenFilter
from ptls.frames.coles.split_strategy import SampleSlices
from ptls.preprocessing.pandas_preprocessor import PandasDataPreprocessor
from ptls.frames.coles import CoLESModule, ColesDataset
from ptls.nn.seq_encoder import RnnSeqEncoder
from ptls.nn.trx_encoder import TrxEncoder


def test_train_inference():
    source_data = pd.read_csv(Path(__file__).parent / "age-transactions.csv")

    preprocessor = PandasDataPreprocessor(
        col_id='client_id',
        col_event_time='trans_date',
        event_time_transformation='none',
        cols_category=["trans_date", "small_group"],
        cols_numerical=["amount_rur"],
    )

    dataset = preprocessor.fit_transform(source_data)
    train, test = train_test_split(dataset, test_size=0.4, random_state=42)
    train_ds = ColesDataset(
        data=MemoryMapDataset(data=train, i_filters=[SeqLenFilter(min_seq_len=5)]),
        splitter=SampleSlices(
            split_count=3,
            cnt_min=3,
            cnt_max=20,
        ),
    )
    train_dl = torch.utils.data.DataLoader(
        train_ds,
        collate_fn=train_ds.collate_fn,
        shuffle=True,
        num_workers=0,
        batch_size=4
    )

    seq_encoder = RnnSeqEncoder(
        trx_encoder=TrxEncoder(
            embeddings={k: {'in': v, 'out': 16} for k, v in preprocessor.get_category_dictionary_sizes().items()},
            numeric_values={},
            embeddings_noise=0.003,
        ),
        hidden_size=16,
        type='gru',
    )
    model = CoLESModule(
        seq_encoder=seq_encoder,
        optimizer_partial=partial(torch.optim.Adam),
        lr_scheduler_partial=partial(torch.optim.lr_scheduler.StepLR, step_size=1, gamma=1.0),
    )

    trainer = pl.Trainer(
        max_epochs=1,
        gpus=0 if torch.cuda.is_available() else 0,
        logger=False
    )
    trainer.fit(model, train_dl)

    test_dl = inference_data_loader(test, num_workers=0, batch_size=4)

    trainer.predict(model, test_dl)



from ptls.data_load.iterable_processing.iterable_shuffle import IterableShuffle
import numpy as np


def get_data_list():
    return [{'uid': i, 'mcc': np.arange(i // 2 + 10)} for i in range(10)]


def get_data_gen():
    for i in  [{'uid': i, 'mcc': np.arange(i // 2 + 10)} for i in range(10)]:
        yield i


def get_data_with_target():
    return [({'uid': i, 'mcc': np.arange(i // 2 + 10)}, i) for i in range(10)]


def test_large_buffer_list():
    i_filter = IterableShuffle(buffer_size=20)
    data = i_filter(get_data_list())
    data = [rec['uid'] for rec in data]
    print(data)
    print(np.array(data) - np.arange(len(data)))
    assert sum(data) == 45
    data = sorted(data)
    assert data == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]


def test_gen():
    i_filter = IterableShuffle(buffer_size=20)
    data = i_filter(get_data_gen())
    data = [rec['uid'] for rec in data]
    print(data)
    print(np.array(data) - np.arange(len(data)))
    assert sum(data) == 45
    data = sorted(data)
    assert data == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]


def test_small_buffer_list():
    i_filter = IterableShuffle(buffer_size=2)
    data = i_filter(get_data_list())
    data = [rec['uid'] for rec in data]
    print(data)
    print(np.array(data) - np.arange(len(data)))
    assert sum(data) == 45
    data = sorted(data)
    assert data == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]


def test_small_buffer_with_target():
    i_filter = IterableShuffle(buffer_size=2)
    data = i_filter(get_data_with_target())
    data = [rec[0]['uid'] for rec in data]
    print(data)
    print(np.array(data) - np.arange(len(data)))
    assert sum(data) == 45
    data = sorted(data)
    assert data == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]



import json
import logging

import hydra
import numpy as np
import pytorch_lightning as pl
from omegaconf import DictConfig
from pytorch_lightning.loggers import TensorBoardLogger

logger = logging.getLogger(__name__)


def fold_fit_test(conf, fold_id):
    conf['fold_id'] = fold_id

    model = hydra.utils.instantiate(conf.pl_module)
    model.seq_encoder.is_reduce_sequence = True
    dm = hydra.utils.instantiate(conf.data_module)

    _trainer_params = conf.trainer
    _trainer_params_additional = {}
    if 'logger_name' in conf:
        _trainer_params_additional['logger'] = TensorBoardLogger(
            save_dir='lightning_logs',
            name=conf.get('logger_name'),
        )
    trainer = pl.Trainer(**_trainer_params, **_trainer_params_additional)
    trainer.fit(model, datamodule=dm)

    valid_metrics = {name: float(mf.compute().item()) for name, mf in model.valid_metrics.items()}
    trainer.test(model=model, dataloaders=dm.test_dataloader(), ckpt_path="best", verbose=False)
    test_metrics = {name: float(mf.compute().item()) for name, mf in model.test_metrics.items()}

    print(', '.join([f'valid_{name}: {v:.4f}' for name, v in valid_metrics.items()]))
    print(', '.join([f' test_{name}: {v:.4f}' for name, v in test_metrics.items()]))

    return {
        "fold_id": fold_id,
        "model_name": conf.embedding_validation_results.model_name,
        "feature_name": conf.embedding_validation_results.feature_name,
        'scores_valid': valid_metrics,
        'scores_test': test_metrics,
    }


@hydra.main(version_base=None)
def main(conf: DictConfig):
    if 'seed_everything' in conf:
        pl.seed_everything(conf.seed_everything)

    fold_list = hydra.utils.instantiate(conf.fold_list)

    results = []
    for fold_id in fold_list:
        logger.info(f'==== Fold [{fold_id}/{len(fold_list)}] fit-test start ====')
        result = fold_fit_test(conf, fold_id)
        results.append(result)

    stats_file = conf.embedding_validation_results.output_path
    if stats_file is not None:
        with open(stats_file, 'w') as f:
            json.dump(results, f, indent=2)
        logger.info(f'Embeddings validation scores saved to "{stats_file}"')

    for name in results[0]['scores_valid'].keys():
        valid_scores = np.array([x['scores_valid'][name]for x in results])
        test_scores = np.array([x['scores_test'][name] for x in results])
        print(f'Valid {name:10}: {valid_scores.mean():.3f} [{", ".join(f"{t:.3f}" for t in valid_scores)}], '
              f'Test  {name:10}: {test_scores.mean():.3f} [{", ".join(f"{t:.3f}" for t in test_scores)}]')


if __name__ == '__main__':
    main()



import torch

from ptls.nn.seq_encoder.rnn_encoder import RnnEncoder
from ptls.nn.seq_encoder.utils import PerTransHead, scoring_head
from ptls.nn.trx_encoder import TrxEncoder
from ptls.nn.trx_encoder.trx_mean_encoder import TrxMeanEncoder


def trx_avg_model(params):
    p = TrxEncoder(params.trx_encoder)
    h = PerTransHead(p.output_size)
    m = torch.nn.Sequential(p, h, torch.nn.Sigmoid())
    return m


def trx_avg2_model(params):
    p = TrxMeanEncoder(params.trx_encoder)
    h = scoring_head(TrxMeanEncoder.output_size(params.trx_encoder), params.head)
    m = torch.nn.Sequential(p, h)
    return m


def rnn_model(params):
    p = TrxEncoder(**params.trx_encoder)
    e = RnnEncoder(p.output_size, **params.rnn)
    h = scoring_head(
        input_size=params.rnn.hidden_size * (2 if params.rnn.bidir else 1),
        params=params.head
    )

    m = torch.nn.Sequential(p, e, h)
    return m



import logging

import hydra
import pytorch_lightning as pl
import torch
from omegaconf import DictConfig
from pytorch_lightning.callbacks import LearningRateMonitor
from pytorch_lightning.callbacks import ModelCheckpoint
from pytorch_lightning.loggers import TensorBoardLogger

logger = logging.getLogger(__name__)


@hydra.main(version_base='1.2', config_path=None)
def main(conf: DictConfig):
    if 'seed_everything' in conf:
        pl.seed_everything(conf.seed_everything)

    model = hydra.utils.instantiate(conf.pl_module)
    dm = hydra.utils.instantiate(conf.data_module)

    _trainer_params = conf.trainer
    _trainer_params_additional = {}
    _use_best_epoch = _trainer_params.get('use_best_epoch', False)

    if 'callbacks' in _trainer_params:
        logger.warning(f'Overwrite `trainer.callbacks`, was "{_trainer_params.checkpoint_callback}"')
    _trainer_params_callbacks = []

    if _use_best_epoch:
        checkpoint_callback = ModelCheckpoint(monitor=model.metric_name, mode='max')
        logger.info(f'Create ModelCheckpoint callback with monitor="{model.metric_name}"')
        _trainer_params_callbacks.append(checkpoint_callback)

    if _trainer_params.get('checkpoints_every_n_val_epochs', False):
        every_n_val_epochs = _trainer_params.checkpoints_every_n_val_epochs
        checkpoint_callback = ModelCheckpoint(every_n_epochs=every_n_val_epochs, save_top_k=-1)
        logger.info(f'Create ModelCheckpoint callback every_n_epochs ="{every_n_val_epochs}"')
        _trainer_params_callbacks.append(checkpoint_callback)

        if 'checkpoint_callback' in _trainer_params:
            del _trainer_params.checkpoint_callback
        if 'enable_checkpointing' in _trainer_params:
            del _trainer_params.enable_checkpointing
        del _trainer_params.checkpoints_every_n_val_epochs

    if 'logger_name' in conf:
        _trainer_params_additional['logger'] = TensorBoardLogger(
            save_dir='lightning_logs',
            name=conf.get('logger_name'),
        )

    lr_monitor = LearningRateMonitor(logging_interval='step')
    _trainer_params_callbacks.append(lr_monitor)

    if len(_trainer_params_callbacks) > 0:
        _trainer_params_additional['callbacks'] = _trainer_params_callbacks
    trainer = pl.Trainer(**_trainer_params, **_trainer_params_additional)
    trainer.fit(model, dm)

    if 'model_path' in conf:
        if _use_best_epoch:
            # from shutil import copyfile
            # copyfile(checkpoint_callback.best_model_path, conf.model_path)
            model.load_from_checkpoint(checkpoint_callback.best_model_path)
            torch.save(model.seq_encoder, conf.model_path)
            logging.info(f'Best model stored in "{checkpoint_callback.best_model_path}" '
                         f'and copied to "{conf.model_path}"')
        else:
            torch.save(model.seq_encoder, conf.model_path)
            logger.info(f'Model weights saved to "{conf.model_path}"')


if __name__ == '__main__':
    main()



import logging
import os
from collections import defaultdict, Counter
from itertools import chain

import numpy as np
import torch
from torch.utils.data import DataLoader

from ptls.data_load import read_pyarrow_file

from pydoc import locate

logger = logging.getLogger(__name__)


class PartitionedDataFiles:
    """PartitionedDataFiles take a structure from partitioned parquet file.
    Structure is an indexes of main partitions (date) and subpartitions (hashes)

    File structure:
    mon=YYYY-MM-DD/
        hash_id=xxx/
            *.parquet

    File structure example:
    data/
        mon=2014-01-31/
            hash_id=1/
                part1.parquet
                part2.parquet
                ...
            hash_id=2/
                part1.parquet
                part2.parquet
                ...
            ...

        mon=2014-02-28/
            hash_id=1/
                ...
            hash_id=2/
                ...


    Data is a matrix: `mon` x `hash_id`

    `hash_id` is a value of `customer_id`. So if we have to collect all data for specific `customer_id`,
    we should calculate `hash_id = H` and get data from all month from `hash_id=H` subpartitions.
    """

    def __init__(self, data_path, dt_start=None, dt_end=None, dt_dtype='datetime64'):
        self._data_path = data_path
        self.to_dtype = locate(dt_dtype)
        self.dt_start = self.to_dtype(dt_start) if dt_start else None
        self.dt_end = self.to_dtype(dt_end) if dt_end else None

        self._dt_parts = None
        self._hash_parts = None

        self._index_files()

    @property
    def data_path(self):
        return self._data_path

    @property
    def dt_parts(self):
        return self._dt_parts

    @property
    def hash_parts(self):
        return self._hash_parts

    def _index_files(self):
        dt_parts = os.listdir(self._data_path)
        dt_parts = [x for x in dt_parts if x.find('=') > 1]
        dt_dates = np.array([x.split('=')[1] for x in dt_parts]).astype(self.to_dtype)

        dt_dates_sort_ix = np.argsort(dt_dates)
        dt_parts = [dt_parts[i] for i in dt_dates_sort_ix]
        dt_dates = dt_dates[dt_dates_sort_ix]

        mask = np.ones_like(dt_dates, dtype=np.bool)
        if self.dt_start is not None:
            mask &= dt_dates >= self.dt_start
        if self.dt_end is not None:
            mask &= dt_dates <= self.dt_end
        self._dt_parts = [x for x, f in zip(dt_parts, mask) if f]

        if len(self._dt_parts) == 0:
            raise AttributeError(f'Empty file list for "{self._data_path}" '
                                 f'in interval [{self.dt_start}, {self.dt_end}]')

        hash_parts = sorted(os.listdir(os.path.join(self._data_path, self._dt_parts[0])))
        for dt_part in self._dt_parts[1:]:
            hash_part_2 = sorted(os.listdir(os.path.join(self._data_path, dt_part)))
            if hash_part_2 != hash_parts:
                raise AttributeError(f'Hash sets in "{self._dt_parts[0]}" and "{dt_part}" are different')

        self._hash_parts = hash_parts


class PartitionedDataset(torch.utils.data.IterableDataset):
    """Lazy (IterableDataset) load from partitioned parquet file

    File structure:
    mon=YYYY-MM-DD/
        hash_id=xxx/
            *.parquet

    File structure example:
    data/
        mon=2014-01-31/
            hash_id=1/
                part1.parquet
                part2.parquet
                ...
            hash_id=2/
                part1.parquet
                part2.parquet
                ...
            ...

        mon=2014-02-28/
            hash_id=1/
                ...
            hash_id=2/
                ...


    Data is a matrix: `mon` x `hash_id`

    `hash_id` is a value of `customer_id`. So if we have to collect all data for specific `customer_id`,
    we should calculate `hash_id = H` and get data from all month from `hash_id=H` subpartitions.
    """
    def __init__(self, data_path, dt_parts, hash_parts,
                 col_id, file_level_processing=None, post_processing=None,
                 shuffle_files=False, cache_schema=True, shuffle_seed=42):
        self.data_path = data_path
        self.dt_parts = dt_parts
        self.hash_parts = hash_parts
        self.col_id = col_id
        self.file_level_processing = file_level_processing
        self.post_processing = post_processing
        self.shuffle_files = shuffle_files
        self.cache_schema = cache_schema
        self.shuffle_seed = shuffle_seed
        self.rs = None

        self._worker_id = None
        self._num_workers = None
        self._shuffle_seed = None
        self._schema = None

    def _get_my_hashes(self):
        files = [(i, name) for i, name in enumerate(self.hash_parts)]
        my_files = [name for i, name in files if i % self._num_workers == self._worker_id]

        return my_files

    def _init_worker(self):
        worker_info = torch.utils.data.get_worker_info()
        if worker_info is None:  # single-process data loading, return the full iterator
            self._worker_id = 0
            self._num_workers = 1
            self._shuffle_seed = self.shuffle_seed
        else:  # in a worker process
            self._worker_id = worker_info.id
            self._num_workers = worker_info.num_workers
            self._shuffle_seed = worker_info.seed
        logger.debug(f'Started [{self._worker_id:02d}/{self._num_workers:02d}]')

    def __iter__(self):
        self._init_worker()

        my_hashes = self._get_my_hashes()

        if self.shuffle_files:
            rs = np.random.RandomState(self._shuffle_seed % 2**32)
            rs.shuffle(my_hashes)

        logger.debug(f'Iter [{self._worker_id:02d}/{self._num_workers:02d}]: {my_hashes}')
        gen = chain(*[self.iter_hash(name) for name in my_hashes])
        if self.post_processing is not None:
            gen = self.post_processing(gen)
        return gen

    def iter_hash(self, hash_part_name):
        """Read all clients for all dt_parts from hash and join dt_parts into a single record

        :param hash_part_name:
        :return:
        """
        logger.debug(f'[{self._worker_id:2d}/{self._num_workers}] Start iter hash {hash_part_name}')
        client_partitioned_data = defaultdict(list)
        for dt_part in self.dt_parts:
            path = os.path.join(self.data_path, dt_part, hash_part_name)
            gen = self.iter_file(path)
            if self.file_level_processing is not None:
                gen = self.file_level_processing(gen)
            for features in gen:
                _id = features[self.col_id]
                cl_feature_list = client_partitioned_data[_id]
                cl_feature_list.append(features)

        for cli_id, features in client_partitioned_data.items():
            yield self.join_features(features)

    def iter_file(self, file_name):
        """

        :param file_name:
        :return: [(customer_id, features)]
        """
        logger.debug(f'[{self._worker_id}/{self._num_workers}] Iter file "{file_name}"')
        for rec in read_pyarrow_file(file_name, use_threads=True):
            yield rec

    def join_features(self, features):
        schema = self.get_schema(features)

        joined_record = {k: self.join_key(k, dtype, [v[k] for v in features]) for k, dtype in schema.items()}
        return joined_record

    def get_schema(self, rows):
        if self._schema is not None and self.cache_schema:
            return self._schema

        schema = set((k, type(v)) for row in rows for k, v in row.items())
        duplicated_keys = [k for k, v in Counter(k for k, v in schema).items() if v > 1]
        if len(duplicated_keys) > 0:
            raise AttributeError(f'Found duplicated_keys {duplicated_keys} '
                                 f'with types {[(k, v) for k, v in schema if k in duplicated_keys]}')
        self._schema = {k: v for k, v in schema}
        return self._schema

    def join_key(self, key, dtype, values):
        if dtype is np.ndarray:
            x = np.concatenate(values)
            if x.dtype.kind in ('i', 'f'):
                return torch.from_numpy(x)
            return x
        elif key == self.col_id:
            return values[0]
        else:
            return values



from ptls.data_load.iterable_processing.target_join import TargetJoin


def get_features_only():
    return [
        {'id': 1},
        {'id': 2},
        {'id': 3},
        {'id': 4},
    ]


def get_features_with_incorrect_target():
    return [
        ({'id': 1}, 0),
        ({'id': 2}, 1),
        ({'id': 3}, 0),
        ({'id': 4}, 1),
    ]

def get_targets():
    return {1: 0, 2: 0, 3: 1, 4: 1}


def test_join():
    i_filter = TargetJoin('id', get_targets())
    data = i_filter(get_features_only())
    data = [y for x, y in data]
    assert data == [0, 0, 1, 1]


def test_join_and_replace():
    i_filter = TargetJoin('id', get_targets())
    orig_data = get_features_with_incorrect_target()

    data = orig_data
    data = [y for x, y in data]
    assert data == [0, 1, 0, 1]  # without replace

    data = i_filter(orig_data)
    data = [y for x, y in data]
    assert data == [0, 0, 1, 1]  # with replace





import torch

from ptls.data_load.padded_batch import PaddedBatch
from ptls.nn.seq_encoder.statistics_encoder import StatisticsEncoder


def get_data():
    payload = {'amount': torch.arange(4*10).view(4, 10).float(),
               'event_time': torch.arange(4*10).view(4, 10).float(),
               'mcc_code': torch.arange(4*10).view(4, 10),
               'tr_type': torch.arange(4*10).view(4, 10)
              }
    return PaddedBatch(
                       payload=payload,
                       length=torch.tensor([4, 2, 6, 8])
                      )

def test_shape():
    config = {
        'top_negative_trx': list(range(0, 20)),
        'top_positive_trx': list(range(20, 40)),       
        'category_names': ['mcc_code', 'tr_type'],
              'numeric_values': ['amount'],
              'category_max_size' : {
                  'mcc_code': 200,
                  'tr_type': 100
              }
    }

    eps = 1e-2

    model = StatisticsEncoder(**config)

    x = get_data()

    out = model(x)
    
    assert isinstance(out, tuple) and len(out) == 4
    assert (abs(out[0] -  torch.Tensor([[-16.1181],
                                        [-16.1181],
                                        [-16.1181],
                                        [-16.1181]])) < torch.zeros((4, 1)) + eps).all()
    assert out[1].shape == torch.Size([4, 6]) and out[1][1][4] == 0 and out[1][3][1] == 0
    assert out[2].shape == torch.Size([4, 1]) and abs(out[2][0].item() - 3.3030) < eps
    assert out[3].shape == torch.Size([4, 6]) and abs(out[3][2][3].item() - 0.0858) < eps



import pandas as pd


class ColTransformerPandasMixin:
    def check_is_col_exists(self, x: pd.DataFrame):
        if self.col_name_original not in x.columns:
            raise AttributeError(f'col_name_original="{self.col_name_original}" not in source dataframe. '
                                 f'Found {x.columns}')

    def drop_original_col(self, x: pd.DataFrame):
        if self.col_name_original != self.col_name_target and self.is_drop_original_col:
            x = x.drop(columns=self.col_name_original)
        return x

    @staticmethod
    def attach_column(df: pd.DataFrame, s: pd.Series):
        new_col_name = s.name
        cols = [col for col in df.columns if col != new_col_name]
        return pd.concat([df[cols], s], axis=1)



from collections import OrderedDict

import numpy as np
import pytest
import torch

from ptls.data_load.feature_dict import FeatureDict

# FeatureDict.is_seq_feature
def test_is_seq_feature_int():
    x = 1
    assert not FeatureDict.is_seq_feature('bin', x)


def test_is_seq_feature_target_int():
    x = 1
    assert not FeatureDict.is_seq_feature('target', x)


def test_is_seq_feature_float():
    x = 1.0
    assert not FeatureDict.is_seq_feature('amount', x)


def test_is_seq_feature_str():
    x = 'user_001'
    assert not FeatureDict.is_seq_feature('user_id', x)


def test_is_seq_feature_list():
    x = [1, 2, 3, 4]
    assert not FeatureDict.is_seq_feature('mcc', x)


def test_is_seq_feature_array():
    x = np.array([1, 2, 3, 4])
    assert FeatureDict.is_seq_feature('mcc', x)


def test_is_seq_feature_event_time():
    x = None
    assert FeatureDict.is_seq_feature('event_time', x)


def test_is_seq_feature_tensor():
    x = torch.Tensor([1, 2, 3, 4])
    assert FeatureDict.is_seq_feature('mcc', x)


def test_is_seq_feature_target_array():
    x = torch.Tensor([1, 2, 3, 4])
    assert not FeatureDict.is_seq_feature('target', x)


def test_seq_indexing_seq():
    x = {'mcc': torch.Tensor([1, 2, 3, 4])}
    torch.testing.assert_close(FeatureDict.seq_indexing(x, [1, 2])['mcc'], torch.Tensor([2, 3]))

# FeatureDict.seq_indexing
def test_seq_indexing_scalar():
    x = {
        'bin': 1001,
        'target_bin': 3,
        'pp': 0.10,
        'user_id': 'u_01',
        'l_mcc': [1, 2, 3, 4],
        't_mcc': torch.IntTensor([5, 6, 7, 8, 9]),
        'event_time': torch.linspace(0, 1, 5),
        'target_dist': torch.linspace(0, 1, 11),
    }
    y = FeatureDict.seq_indexing(x, [1, 2])
    assert y['bin'] == 1001
    assert y['target_bin'] == 3
    assert y['pp'] == 0.1
    assert y['user_id'] ==  'u_01'
    assert y['l_mcc'] == [1, 2, 3, 4]
    torch.testing.assert_close(y['t_mcc'], torch.IntTensor([6, 7]))
    torch.testing.assert_close(y['event_time'], torch.FloatTensor([0.25, 0.5]))
    torch.testing.assert_close(y['target_dist'], torch.arange(0, 1.1, 0.1))


# FeatureDict.get_seq_len
def test_seq_len_et():
    x = OrderedDict([
        ('target', torch.Tensor([1, 2, 3, 4, 5])),
        ('bin', 2),
        ('mcc', torch.Tensor([1, 2, 3, 4])),
        ('event_time', torch.Tensor([1, 2, 3])),
    ])
    assert FeatureDict.get_seq_len(x) == 3


def test_seq_len_seq():
    x = OrderedDict([
        ('target', torch.Tensor([1, 2, 3, 4, 5])),
        ('bin', 2),
        ('mcc', torch.Tensor([1, 2, 3, 4])),
        ('_event_time', torch.Tensor([1, 2, 3])),
    ])
    assert FeatureDict.get_seq_len(x) == 4


def test_seq_len_nolen():
    with pytest.raises(StopIteration):
        x = OrderedDict([
            ('target', torch.Tensor([1, 2, 3, 4, 5])),
            ('bin', 2),
        ])
        FeatureDict.get_seq_len(x)



import numpy as np
import torch

from ptls.nn.seq_encoder.utils import get_distributions, transform_inv
from ptls.data_load.padded_batch import PaddedBatch


class StatisticsEncoder(torch.nn.Module):
    def __init__(self,
                 pos=True,
                 neg=True,
                 top_negative_trx=[],
                 top_positive_trx=[],
                 category_names=None,
                 numeric_values=None,
                 category_max_size=None):
        super().__init__()
        self.collect_pos, self.collect_neg = pos, neg
        self.dummy = torch.nn.Linear(1, 1)
        self.negative_items = top_negative_trx
        self.positive_items = top_positive_trx

        self.cat_names = list(category_names)
        self.num_values = list(numeric_values)
        self.cat_max_size = category_max_size

    def forward(self, x: PaddedBatch):
        eps = 1e-7
        tr_type_col = []
        amount_col = []
        for i, row in enumerate(zip(x.payload[self.cat_names[0]], x.payload[self.num_values[0]])):
            tr_type_col += [list(row[0][:x.seq_lens[i].item()].cpu().numpy())]
            amount_col += [list(row[1][:x.seq_lens[i].item()].cpu().numpy())]

        tr_type_col = np.array(tr_type_col, dtype=object)[:, None]
        amount_col = np.array(amount_col, dtype=object)[:, None]
        np_data = np.hstack((tr_type_col, amount_col))

        distr = get_distributions(np_data, 1, 0, self.negative_items, self.positive_items, 5, 0, transform_inv)

        if self.collect_neg:
            sums_of_negative_target = torch.tensor(distr[0])[:, None]
            neg_distribution = torch.tensor(distr[2])
            log_neg_sum = torch.log(torch.abs(sums_of_negative_target + eps))
        if self.collect_pos:
            sums_of_positive_target = torch.tensor(distr[1])[:, None]
            pos_distribution = torch.tensor(distr[3])
            log_pos_sum = torch.log(sums_of_positive_target + eps)

        if self.collect_neg and self.collect_pos:
            return log_neg_sum, neg_distribution, log_pos_sum, pos_distribution
        elif self.collect_neg:
            return log_neg_sum, neg_distribution, 0, 0
        elif self.collect_pos:
            return 0, 0, log_pos_sum, pos_distribution

    @property
    def category_names(self):
        return set(self.cat_names + self.num_values)

    @property
    def category_max_size(self):
        return self.cat_max_size



import torch


class FastTensorDataLoader:
    """
    A DataLoader-like object for a set of tensors that can be much faster than
    TensorDataset + DataLoader because dataloader grabs individual indices of
    the dataset and calls cat (slow).
    Source: https://discuss.pytorch.org/t/dataloader-much-slower-than-manual-batching/27014/6
    """
    def __init__(self, *tensors, batch_size=32, shuffle=False, post_process_func=None):
        """
        Initialize a FastTensorDataLoader.

        :param *tensors: tensors to store. Must have the same length @ dim 0.
        :param batch_size: batch size to load.
        :param shuffle: if True, shuffle the data *in-place* whenever an
            iterator is created out of this object.
        :param post_process_func: if not None apply function to the resulting batch

        :returns: A FastTensorDataLoader.
        """
        assert all(t.shape[0] == tensors[0].shape[0] for t in tensors)

        self.tensors = tensors

        self.dataset_len = self.tensors[0].shape[0]
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.post_process_func = post_process_func

        # Calculate # batches
        n_batches, remainder = divmod(self.dataset_len, self.batch_size)
        if remainder > 0:
            n_batches += 1
        self.n_batches = n_batches

    def __iter__(self):
        if self.shuffle:
            r = torch.randperm(self.dataset_len)
            self.tensors = [t[r] for t in self.tensors]
        self.i = 0
        return self

    def __next__(self):
        if self.i >= self.dataset_len:
            raise StopIteration
        if len(self.tensors) > 1:
            batch = tuple(t[self.i:self.i+self.batch_size] for t in self.tensors)
        else:
            batch = self.tensors[0][self.i:self.i+self.batch_size]

        self.i += self.batch_size

        if self.post_process_func is not None:
            batch = self.post_process_func(batch)

        return batch

    def __len__(self):
        return self.n_batches



import torch
from torch import nn as nn
from torch.nn import functional as F
from ptls.frames.coles.sampling_strategies import MatrixMasker


class SoftmaxLoss(nn.Module):
    """
    Softmax loss.
    """

    def __init__(self, masker=None, eps=1e-6, temperature=0.05):
        super(SoftmaxLoss, self).__init__()
        self.masker = masker
        if masker is None:
            self.masker = MatrixMasker()
        self.eps = eps
        self.temperature = temperature

    def forward(self, embeddings, classes):
        similarities = self.get_sim_matrix(embeddings, embeddings, eps=self.eps)
        similarities /= self.temperature
        log_matrix = (-1)*F.log_softmax(similarities)
        masked_matrix = self.masker.get_masked_matrix(log_matrix, classes)
        loss = masked_matrix/len(similarities)
        return loss.sum()

    @staticmethod
    def get_sim_matrix(a, b, eps):
        a_n, b_n = a.norm(dim=1)[:, None], b.norm(dim=1)[:, None]
        a_norm = a / torch.max(a_n, eps * torch.ones_like(a_n))
        b_norm = b / torch.max(b_n, eps * torch.ones_like(b_n))
        sim_mt = torch.mm(a_norm, b_norm.transpose(0, 1))
        return sim_mt



import numpy as np

from ptls.data_load.iterable_processing.target_empty_filter import TargetEmptyFilter


def get_data():
    return [
        {'uid': 1, 'target': 0},
        {'uid': 2, 'target': 1},
        {'uid': 3, 'target': None},
        {'uid': 4, 'target': np.NaN},
        {'uid': 5, 'target': float('NaN')},
        {'uid': 6, 'target': "class_0"},
     ]


def test_filter():
    i_filter = TargetEmptyFilter('target')
    data = i_filter(get_data())
    data = [x['uid'] for x in data]
    assert data == [1, 2, 6]



import torch
from omegaconf import OmegaConf
from ptls.frames.bert.modules.mlm_module import MLMPretrainModule
from ptls.data_load.padded_batch import PaddedBatch
from ptls.nn.seq_encoder.abs_seq_encoder import AbsSeqEncoder
from ptls.nn import RnnEncoder, TransformerEncoder, LongformerEncoder


def get_config():
    return OmegaConf.create(dict(
        hidden_size=8,
        loss_temperature=1.0,
        neg_count=2,
        total_steps=100,
    ))


def test_neg_ix():
    m = MLMPretrainModule(
        trx_encoder=None,
        seq_encoder=AbsSeqEncoder(),
        **OmegaConf.merge(get_config(), OmegaConf.from_dotlist(["neg_count=4"])),
    )
    mask = torch.Tensor([
        [0, 0, 1, 0],
        [0, 1, 1, 0],
        [1, 0, 0, 0],
        [0, 0, 0, 0],
        [0, 0, 0, 1],
    ]).bool()

    b_ix, neg_ix = m.get_neg_ix(mask)
    assert b_ix.size() == (5, 4)
    assert neg_ix.size() == (5, 4)
    assert mask[b_ix, neg_ix].all()


def test_training_step():
    models = [
        RnnEncoder(input_size=8, hidden_size=8),
        TransformerEncoder(input_size=8),
        LongformerEncoder(input_size=8),

    ]
    for seq_encoder in models:
        m = MLMPretrainModule(
            trx_encoder=lambda x: x,
            seq_encoder=seq_encoder,
            **get_config(),
        )

        x = PaddedBatch(
            torch.randn(16, 9, 8),
            torch.randint(3, 9, (16,)),
        )

        loss = m.training_step(x, 0)
        assert loss.item() is not None



from ptls.data_load.iterable_processing_dataset import IterableProcessingDataset


class TargetMove(IterableProcessingDataset):
    """Deprecated. Only single dict allowed.
    Store target as scalar value in a feature dictionary

    Filter that takes target from sample dict and place it in tuple with dict
    Parameters
     ----------
    target_col : str. Default: 'target'
         Name of target columns in sample dict.
         
    Filter transformation:
    list(input_dict) -> list((input_dict, target_val))
    (in place of list there could be other iterable, for example generator)
    """
    def __init__(self, target_col='target'):
        super().__init__()
        self.target_col = target_col

    def __iter__(self):
        for rec in self._src:
            features = rec[0] if type(rec) is tuple else rec
            target = int(features[self.target_col])
            yield features, target




import torch

from ptls.frames.abs_module import ABSModule
from ptls.frames.cpc.losses.cpc_loss import CPC_Loss
from ptls.frames.cpc.metrics.cpc_accuracy import CpcAccuracy
from ptls.nn.seq_encoder import RnnSeqEncoder
from ptls.data_load.padded_batch import PaddedBatch


class CpcV2Module(ABSModule):
    """Contrastive Predictive Coding ([CPC](https://arxiv.org/abs/1807.03748))

    Original sequence splits at small parts which are encoded by `seq_encoder` with reduce.
    Hidden representation `z` is an embedding vector for small sequence split.
    Next `aggregator` `RnnEncoder` used for `context` calculation from `z`.
    Linear predictors are used to predict next trx embedding by context.
    The loss function tends to make future trx embedding and they predict closer.
    Negative sampling are used to avoid trivial solution.

    Parameters
        seq_encoder:
            Model which calculate embeddings for original raw transaction sequences
            `seq_encoder` is trained by `CoLESModule` to get better representations of input sequences
        head:
            Not used
        loss:
            Keep None. CPCLoss used by default
        aggregator:
            RnnEncoder which are used to calculate context
        validation_metric:
            Keep None. CPCAccuracy used by default
        optimizer_partial:
            optimizer init partial. Network parameters are missed.
        lr_scheduler_partial:
            scheduler init partial. Optimizer are missed.

    """
    def __init__(self, validation_metric=None,
                       seq_encoder=None,
                       head=None,
                       n_negatives=40, n_forward_steps=6,
                       aggregator=None,
                       optimizer_partial=None,
                       lr_scheduler_partial=None):

        self.save_hyperparameters('n_negatives', 'n_forward_steps')

        loss = CPC_Loss(n_negatives=n_negatives, n_forward_steps=n_forward_steps)
        if validation_metric is None:
            validation_metric = CpcAccuracy(loss)

        if seq_encoder is not None and not isinstance(seq_encoder, RnnSeqEncoder):
            raise NotImplementedError(f'Only rnn encoder supported in CpcModule. Found {type(seq_encoder)}')

        super().__init__(validation_metric,
                         seq_encoder,
                         loss,
                         optimizer_partial,
                         lr_scheduler_partial)

        # Define aggregator
        self.agg_rnn = aggregator

        # Define linear transformation for every cpc.n_forward_steps
        self.n_forward_steps = loss.n_forward_steps
        embedding_size = aggregator.hidden_size
        linear_size = seq_encoder.rnn_encoder.hidden_size
        self._linears = torch.nn.ModuleList([torch.nn.Linear(embedding_size, linear_size)
                                             for _ in range(self.n_forward_steps)])


    @property
    def metric_name(self):
        return 'cpc_accuracy'

    @property
    def is_requires_reduced_sequence(self):
        return True

    def shared_step(self, x, y):
        """
        x: list of PaddedBatch,
        where PaddedBatch = (
            payload = {key: [padded tensor] whit shape [batch_size, max_subseq_len]},
            seq_lens = (1 x batch_size)
        )
        """
        split_embeddings = []
        for padded_batch in x:
            split_embeddings.append(self._seq_encoder(padded_batch))

        base_embeddings = PaddedBatch(torch.stack(split_embeddings, dim=1), None)
        context_embeddings = self.agg_rnn(base_embeddings)


        me = []
        for l in self._linears:
            me.append(l(context_embeddings.payload))

        batch_size, max_seq_len, _ = base_embeddings.payload.shape

        mapped_ctx_embeddings = PaddedBatch(
            torch.stack(me, dim=3),
            max_seq_len * torch.ones(batch_size, device=x[0].device)
        )
        return (base_embeddings, context_embeddings, mapped_ctx_embeddings), y

    def validation_step(self, batch, _):
        y_h, y = self.shared_step(batch, None)
        self._validation_metric(y_h, y)

    def training_step(self, batch, _):
        y_h, y = self.shared_step(batch, None)
        loss = self._loss(y_h, y)
        self.log('loss', loss)
        return loss



import torch

from ptls.frames.coles.sampling_strategies import AllTripletSelector, RandomNegativeTripletSelector, \
    HardTripletSelector, SemiHardTripletSelector


def get_data():
    b, c, h = 3, 2, 2  # Batch, num Classes, Hidden size
    x = torch.tensor([[0., 1.],
                      [1., 0.],
                      [1., 1.],
                      [0.1, 0.9],
                      [0.8, 0.2],
                      [1.0, 0.9]])

    y = torch.arange(c).view(-1, 1).expand(c, b).reshape(-1)
    return x, y


def check_triplets(triplets, y):
    # positives from same classes as anchor elements
    assert torch.equal(y[triplets[:, 0]], y[triplets[:, 1]])

    # positives from another classes than anchor elements
    assert torch.all(~torch.eq(y[triplets[:, 0]], y[triplets[:, 2]]))


def test_random_triplet_selector():
    x, y = get_data()
    sampling_strategy = RandomNegativeTripletSelector()

    triplets = sampling_strategy.get_triplets(x, y)
    check_triplets(triplets, y)


def test_all_triplet_selector():
    x, y = get_data()
    sampling_strategy = AllTripletSelector()

    triplets = sampling_strategy.get_triplets(x, y)
    check_triplets(triplets, y)

    true_triplets = torch.LongTensor([[0, 1, 3],
                                      [0, 1, 4],
                                      [0, 1, 5],
                                      [0, 2, 3],
                                      [0, 2, 4],
                                      [0, 2, 5],
                                      [1, 2, 3],
                                      [1, 2, 4],
                                      [1, 2, 5],
                                      [3, 4, 0],
                                      [3, 4, 1],
                                      [3, 4, 2],
                                      [3, 5, 0],
                                      [3, 5, 1],
                                      [3, 5, 2],
                                      [4, 5, 0],
                                      [4, 5, 1],
                                      [4, 5, 2]])

    assert torch.equal(triplets, true_triplets)


def test_hard_triplet_selector():
    x, y = get_data()
    sampling_strategy = HardTripletSelector()

    triplets = sampling_strategy.get_triplets(x, y)
    check_triplets(triplets, y)

    true_triplets = torch.LongTensor([[0, 1, 3],
                                      [0, 2, 3],
                                      [1, 2, 4],
                                      [3, 4, 0],
                                      [3, 5, 0],
                                      [4, 5, 1]])

    assert torch.equal(triplets, true_triplets)


def test_semihard_triplet_selector():
    x, y = get_data()
    sampling_strategy = SemiHardTripletSelector()

    triplets = sampling_strategy.get_triplets(x, y)
    check_triplets(triplets, y)

    true_triplets = torch.LongTensor([[0, 1, 4],
                                      [0, 2, 5],
                                      [1, 2, 3],
                                      [3, 4, 1],
                                      [3, 5, 2],
                                      [4, 5, 2]])

    assert torch.equal(triplets, true_triplets)



from collections import OrderedDict

import numpy as np
import torch

from ptls.data_load.padded_batch import PaddedBatch


class AggFeatureSeqEncoder(torch.nn.Module):
    """Calculates statistics over feature arrays and return them as embedding.

    Result is high dimension non learnable vector.
    Such embedding can be used in busting ML algorithm.

    Statistics are calculated by numerical features, with grouping by category values.

    This seq-encoder haven't TrxEncoder, they consume raw features.

    Parameters
        embeddings:
            dict with categorical feature names.
            Values must be like this `{'in': dictionary_size}`.
            One hot encoding will be applied to these features.
            Output vector size is (dictionary_size,)
        numeric_values:
            dict with numerical feature names.
            Values may contains some options
            numeric_values are multiplied with OHE categories.
            Then statistics are calculated over time dimensions.
        was_logified (bool):
            True - means that original numerical features was log transformed
            AggFeatureSeqEncoder will use `exp` to get original value
        log_scale_factor:
            Use it with `was_logified=True`. value will be multiplied by log_scale_factor before exp.
        is_used_count (bool):
            Use of not count by values
        is_used_mean (bool):
            Use of not mean by values
        is_used_std (bool):
            Use of not std by values
        is_used_min (bool):
            Use of not min by values
        is_used_max (bool):
            Use of not max by values
        use_topk_cnt (int):
            Define the K for topk features calculation. 0 if not used
        distribution_target_task (bool):
            Calc more features
        logify_sum_mean_seqlens (bool):
            True - apply log transform to sequence length

    Example:

    """
    def __init__(self,
                 embeddings=None,
                 numeric_values=None,
                 was_logified=True,
                 log_scale_factor=1,
                 is_used_count=True,
                 is_used_mean=True,
                 is_used_std=True,
                 is_used_min=False,
                 is_used_max=False,
                 use_topk_cnt=0,
                 distribution_target_task=False,
                 logify_sum_mean_seqlens=False,
                 ):

        super().__init__()

        self.numeric_values = OrderedDict(numeric_values.items())
        self.embeddings = OrderedDict(embeddings.items())
        self.was_logified = was_logified
        self.log_scale_factor = log_scale_factor
        self.distribution_target_task = distribution_target_task
        self.logify_sum_mean_seqlens = logify_sum_mean_seqlens

        self.eps = 1e-9

        self.ohe_buffer = {}
        for col_embed, options_embed in self.embeddings.items():
            size = options_embed['in']
            ohe = torch.diag(torch.ones(size))
            self.ohe_buffer[col_embed] = ohe
            self.register_buffer(f'ohe_{col_embed}', ohe)

        self.is_used_count = is_used_count
        self.is_used_mean = is_used_mean
        self.is_used_std = is_used_std
        self.is_used_min = is_used_min
        self.is_used_max = is_used_max
        self.use_topk_cnt = use_topk_cnt

    def forward(self, x: PaddedBatch):
        """
        {
            'cat_i': [B, T]: int
            'num_i': [B, T]: float
        }
        to
        {
            [B, H] where H - is [f1, f2, f3, ... fn]
        }
        :param x:
        :return:
        """

        feature_arrays = x.payload
        device = x.device
        B, T = x.seq_feature_shape
        seq_lens = x.seq_lens.to(device).float()
        # if (seq_lens == 0).any():
        #     raise Exception('seq_lens == 0')

        if (self.logify_sum_mean_seqlens):
            processed = [torch.log(seq_lens.unsqueeze(1))]  # count
        else:
            processed = [seq_lens.unsqueeze(1)]
        cat_processed = []

        for col_num, options_num in self.numeric_values.items():
            # take array with numerical feature and convert it to original scale
            if col_num.strip('"') == '#ones':
                val_orig = torch.ones(B, T, device=device)
            else:
                val_orig = feature_arrays[col_num].float()

            if any((
                    type(options_num) is str and self.was_logified,
                    type(options_num) is dict and options_num.get('was_logified', False),
            )):
                #
                val_orig = torch.expm1(self.log_scale_factor * torch.abs(val_orig)) * torch.sign(val_orig)

            # trans_common_features

            sum_ = val_orig.sum(dim=1).unsqueeze(1)  # sum
            a = torch.clamp(val_orig.pow(2).sum(dim=1) - val_orig.sum(dim=1).pow(2).div(seq_lens + self.eps), min=0.0)
            mean_ = val_orig.sum(dim=1).div(seq_lens + self.eps).unsqueeze(1)  # mean
            std_ = a.div(torch.clamp(seq_lens - 1, min=0.0) + self.eps).pow(0.5).unsqueeze(1)  # std

            if self.distribution_target_task:
                sum_pos = torch.clamp(val_orig, min=0).sum(dim=1).unsqueeze(1)
                processed.append(torch.log(sum_pos + 1))  # log sum positive

                sum_neg = torch.clamp(val_orig, max=0).sum(dim=1).unsqueeze(1)
                processed.append(-1 * torch.log(-sum_neg + 1))  # log sum negative
            elif self.logify_sum_mean_seqlens:
                processed.append(torch.sign(sum_) * torch.log(torch.abs(sum_) + 1))
            else:
                processed.append(sum_)

            if self.logify_sum_mean_seqlens:
                mean_ = torch.sign(mean_) * torch.log(torch.abs(mean_) + 1)

            processed.append(mean_)

            if not self.distribution_target_task:
                processed.append(std_)

            # TODO: percentiles

            # embeddings features (like mcc)
            for col_embed, options_embed in self.embeddings.items():
                ohe = getattr(self, f'ohe_{col_embed}')
                val_embed = feature_arrays[col_embed].long()
                val_embed = val_embed.clamp(0, options_embed['in'] - 1)

                ohe_transform = ohe[val_embed.flatten()].view(*val_embed.size(), -1)  # B, T, size
                m_sum = ohe_transform * val_orig.unsqueeze(-1)  # B, T, size

                # counts over val_embed
                mask = (1.0 - ohe[0]).unsqueeze(0)  # 0, 1, 1, 1, ..., 1

                e_cnt = ohe_transform.sum(dim=1) * mask
                if self.is_used_count:
                    processed.append(e_cnt)

                # sum over val_embed
                if self.is_used_mean:
                    e_sum = m_sum.sum(dim=1)
                    e_mean = e_sum.div(e_cnt + 1e-9)
                    processed.append(e_mean)

                if self.is_used_std:
                    a = torch.clamp(m_sum.pow(2).sum(dim=1) - m_sum.sum(dim=1).pow(2).div(e_cnt + 1e-9), min=0.0)
                    e_std = a.div(torch.clamp(e_cnt - 1, min=0) + 1e-9).pow(0.5)
                    processed.append(e_std)

                if self.is_used_min:
                    min_ = m_sum.masked_fill(~x.seq_len_mask.bool().unsqueeze(2), np.float32('inf')).min(dim=1).values
                    processed.append(min_)

                if self.is_used_max:
                    max_ = m_sum.masked_fill(~x.seq_len_mask.bool().unsqueeze(2), np.float32('-inf')).max(dim=1).values
                    processed.append(max_)

        # n_unique and top_k
        for col_embed, options_embed in self.embeddings.items():
            ohe = getattr(self, f'ohe_{col_embed}')
            val_embed = feature_arrays[col_embed].long()
            val_embed = val_embed.clamp(0, options_embed['in'] - 1)

            ohe_transform = ohe[val_embed.flatten()].view(*val_embed.size(), -1)  # B, T, size

            # counts over val_embed
            mask = (1.0 - ohe[0]).unsqueeze(0)  # 0, 1, 1, 1, ..., 1
            e_cnt = ohe_transform.sum(dim=1) * mask

            processed.append(e_cnt.gt(0.0).float().sum(dim=1, keepdim=True))

            if self.use_topk_cnt > 0:
                cat_processed.append(torch.topk(e_cnt, self.use_topk_cnt, dim=1)[1])

        for i, t in enumerate(processed):
            if torch.isnan(t).any():
                raise Exception(f'nan in {i}')

        out = torch.cat(processed + cat_processed, 1)

        return out

    @property
    def embedding_size(self):
        numeric_values = self.numeric_values
        embeddings = self.embeddings

        e_sizes = [options_embed['in'] for col_embed, options_embed in embeddings.items()]

        out_size = 1

        n_features = sum([int(v) for v in [
            self.is_used_count, self.is_used_mean, self.is_used_std, self.is_used_min, self.is_used_max]])
        out_size += len(numeric_values) * (3 + n_features * sum(e_sizes)) + len(embeddings)
        out_size += len(e_sizes) * self.use_topk_cnt
        return out_size

    @property
    def cat_output_size(self):
        e_sizes = [options_embed['in'] for col_embed, options_embed in self.embeddings.items()]
        return len(e_sizes) * self.use_topk_cnt

    @property
    def category_names(self):
        return set([field_name for field_name in self.embeddings.keys()] +
                   [value_name for value_name in self.numeric_values.keys()]
                   )

    @property
    def category_max_size(self):
        return {k: v['in'] for k, v in self.embeddings.items()}



import torch
import logging

logger = logging.getLogger(__name__)


class PersistDataset(torch.utils.data.Dataset):
    def __init__(self, data):
        self.data = [rec for rec in data]
        logger.info(f'Loaded {len(self.data)} records')

    def __len__(self):
        return len(self.data)

    def __getitem__(self, item):
        return self.data[item]



import torch
import torch.nn.functional as F
from torch import nn
import numpy as np
from ptls.data_load.padded_batch import PaddedBatch


def cross_entropy(pred, soft_targets):
    logsoftmax = torch.nn.LogSoftmax(dim=1)
    device = pred.device
    return torch.mean(torch.sum(-soft_targets.to(device) * logsoftmax(pred), 1))


def kl(pred, soft_targets):
    eps = 1e-7
    softmax = torch.nn.Softmax(dim=1)
    device = pred.device
    return torch.mean(torch.sum(soft_targets.to(device) * torch.log(soft_targets.to(device) / (softmax(pred) + eps) + eps), 1))


def mse_loss(pred, actual):
    device = pred.device
    return torch.mean((pred - actual.to(device))**2)


def mape_metric(pred, actual):
    eps = 1
    device = pred.device
    return torch.mean((actual.to(device) - pred).abs() / (actual.to(device).abs() + eps))

def r_squared(pred, actual):
    device = pred.device
    return 1 - torch.sum((actual.to(device) - pred)**2) \
                         / torch.sum((actual.to(device) - torch.mean(actual.to(device)))**2)


class PairwiseMarginRankingLoss(nn.Module):
    def __init__(self, margin=0.0, size_average=None, reduce=None, reduction='mean'):
        """
        Pairwise Margin Ranking Loss. All setted parameters redirected to nn.MarginRankingLoss.
        All the difference is that pairs automatically generated for margin ranking loss.
        All possible pairs of different class are generated.
        """
        super().__init__()
        self.margin_loss = nn.MarginRankingLoss(margin, size_average, reduce, reduction)

    def forward(self, prediction, label):
        """
        Get pairwise margin ranking loss.
        :param prediction: tensor of shape Bx1 of predicted probabilities
        :param label: tensor of shape Bx1 of true labels for pair generation
        """

        # positive-negative selectors
        mask_0 = label == 0
        mask_1 = label == 1

        # selected predictions
        pred_0 = torch.masked_select(prediction, mask_0)
        pred_1 = torch.masked_select(prediction, mask_1)
        pred_1_n = pred_1.size()[0]
        pred_0_n = pred_0.size()[0]

        if pred_1_n > 0 and pred_0_n:
            # create pairs
            pred_00 = pred_0.unsqueeze(0).repeat(1, pred_1_n)
            pred_11 = pred_1.unsqueeze(1).repeat(1, pred_0_n).view(pred_00.size())
            out01 = -1 * torch.ones(pred_1_n*pred_0_n).to(prediction.device)

            return self.margin_loss(pred_00.view(-1), pred_11.view(-1), out01)
        else:
            return torch.sum(prediction) * 0.0


class MultiLoss(nn.Module):
    """Works like `ptls.contrastive_learning.losses.complex_loss.ComplexLoss`

    """
    def __init__(self, losses):
        super().__init__()
        self.losses = nn.ModuleList(losses)

    def forward(self, pred, true):
        loss = 0
        for weight, criterion in self.losses:
            loss = weight * criterion(pred, true) + loss

        return loss

class TransactionSumLoss(nn.Module):
    def __init__(self, n_variables_to_predict):
        super().__init__()
        self.bce_with_logits = nn.BCEWithLogitsLoss()
        self.n_variables_to_predict = n_variables_to_predict

    def forward(self, pred, true):
        #the 0th element is not used, because it is reserved for the total sum of transactions
        #which is not predicted now
        loss = self.bce_with_logits(pred[:,1:self.n_variables_to_predict], true[:,1:self.n_variables_to_predict])
        return loss


class AllStateLoss(nn.Module):
    def __init__(self, point_loss):
        super().__init__()
        self.loss = point_loss

    def forward(self, pred: PaddedBatch, true):
        y = torch.cat([torch.Tensor([yb] * length) for yb, length in zip(true, pred.seq_lens)])
        weights = torch.cat([torch.arange(1, length + 1) / length for length in pred.seq_lens])

        loss = self.loss(pred, y, weights)

        return loss


class BCELoss(nn.Module):
    def __init__(self):
        super().__init__()
        self.loss = nn.BCELoss()

    def forward(self, pred, true):
        return self.loss(pred.float(), true.float())


class MSELoss(nn.Module):
    def __init__(self):
        super().__init__()
        self.loss = nn.MSELoss()

    def forward(self, pred, true):
        return self.loss(pred.float(), true.float())


class PseudoLabeledLoss(nn.Module):
    def __init__(self, loss, pl_threshold=0.5, unlabeled_weight=1.):
        super().__init__()
        self.loss = loss
        self.pl_threshold = pl_threshold
        self.unlabeled_weight = unlabeled_weight

    def forward(self, pred, true):
        label_pred, unlabel_pred = pred['labeled'], pred['unlabeled']
        if isinstance(self.loss, nn.NLLLoss):
            pseudo_labels = torch.argmax(unlabel_pred.detach(), 1)
        elif isinstance(self.loss, BCELoss):
            pseudo_labels = (unlabel_pred.detach() > 0.5).type(torch.int64)
        else:
            raise Exception(f'unknown loss type: {self.loss}')

        # mask pseudo_labels, with confidence > pl_threshold
        if isinstance(self.loss, nn.NLLLoss):
            probs = torch.exp(unlabel_pred.detach())
            mask = (probs.max(1)[0] > self.pl_threshold)
        elif isinstance(self.loss, BCELoss):
            probs = unlabel_pred.detach()
            mask = abs(probs - (1 - pseudo_labels)) > self.pl_threshold
        else:
            mask = torch.ones(unlabel_pred.shape[0]).bool()

        Lloss = self.loss(label_pred, true)
        if mask.sum()==0:
            return Lloss
        else:
            Uloss = self.unlabeled_weight * self.loss(unlabel_pred[mask], pseudo_labels[mask])
            return (Lloss + Uloss) / (1 + self.unlabeled_weight)


class UnsupervisedTabNetLoss(nn.Module):
    def __init__(self, eps=1e-9):
        super().__init__()
        self.eps = eps

    def forward(self, output, obf_vars):
        embedded_x, y_pred = output
        errors = y_pred - embedded_x
        reconstruction_errors = torch.mul(errors, obf_vars) ** 2
        batch_stds = torch.std(embedded_x, dim=0) ** 2 + self.eps
        features_loss = torch.matmul(reconstruction_errors, 1 / batch_stds)
        # compute the number of obfuscated variables to reconstruct
        nb_reconstructed_variables = torch.sum(obf_vars, dim=1)
        # take the mean of the reconstructed variable errors
        features_loss = features_loss / (nb_reconstructed_variables + self.eps)
        # the mean per batch
        loss = torch.mean(features_loss)
        return loss


class DistributionTargetsLoss(nn.Module):
    def __init__(self, mult1=3, mult2=0.167, mult3=1, mult4=1):
        super().__init__()
        self.mults = [mult1, mult2, mult3, mult4]

    def forward(self, pred, true):
        log_sum_truth_neg = np.log(np.abs(true['neg_sum'].astype(np.float)) + 1)[:, None] if isinstance(true['neg_sum'], np.ndarray) else 0
        distribution_truth_neg = np.array(true['neg_distribution'].tolist()) if isinstance(true['neg_sum'], np.ndarray) else 0
        log_sum_truth_pos = np.log(np.abs(true['pos_sum'].astype(np.float)) + 1)[:, None]
        distribution_truth_pos = np.array(true['pos_distribution'].tolist())

        log_sum_hat_neg, distribution_hat_neg = pred['neg_sum'], pred['neg_distribution']
        log_sum_hat_pos, distribution_hat_pos = pred['pos_sum'], pred['pos_distribution']

        device = log_sum_hat_pos.device
        loss_sum_pos = mse_loss(log_sum_hat_pos, torch.tensor(log_sum_truth_pos, device=device).float())
        loss_distr_pos = cross_entropy(distribution_hat_pos, torch.tensor(distribution_truth_pos, device=device))
        if isinstance(true['neg_sum'], np.ndarray):
            loss_sum_neg = mse_loss(log_sum_hat_neg, torch.tensor(log_sum_truth_neg, device=device).float())
            loss_distr_neg = cross_entropy(distribution_hat_neg, torch.tensor(distribution_truth_neg, device=device))

        loss = loss_sum_neg * self.mults[0] + loss_sum_pos * self.mults[1] + loss_distr_neg * self.mults[2] + loss_distr_pos * self.mults[3] \
                    if isinstance(true['neg_sum'], np.ndarray) else loss_sum_pos * self.mults[1] + loss_distr_pos * self.mults[3]
        return loss.float()


class ZILNLoss(nn.Module):
    """
    Zero-inflated lognormal (ZILN) loss adapted for multinomial target with K categories.
    Please cite [https://arxiv.org/abs/1912.07753] and [https://github.com/google/lifetime_value].
    Defaults to MSE loss for 1D-input and lognormal loss for 2D-input.

    Parameters
    ----------
    pred: tensor of shape (B) or (B, K')
    target: tensor of shape (B) or (B, K)
    """
    def __init__(self):
        super().__init__()
        self.eps = 1e-8

    def forward(self, pred, target):
        t = target if target.dim() == 1 else target.sum(dim=1)
        if pred.dim() == 1:
            return torch.mean((pred - t).square())
        sigma = F.softplus(pred[:, 1])
        loss = (sigma * t + self.eps).log() + ((t + self.eps).log() - pred[:, 0]).square() / (2 * sigma.square() + self.eps)
        if pred.shape[1] == 2:
            pass
        elif pred.shape[1] == 3:
            loss = torch.where(t > 0, loss - F.logsigmoid(pred[:, 2]), -F.logsigmoid(-pred[:, 2]))
        elif pred.shape[1] == target.shape[1] + 1:
            dist = torch.cat(((t == 0).float().unsqueeze(-1), target), dim=1)
            loss = -dist.mul(F.log_softmax(pred, dim=1)).sum(dim=1)
        elif pred.shape[1] == target.shape[1] + 3:
            log_prob = F.log_softmax(pred[:, 2:], dim=1)
            loss = torch.where(t > 0, loss - target.mul(log_prob[:, 1:]).sum(dim=1), -log_prob[:, 0])
        else:
            raise Exception(f"{self.__class__} got incorrect input sizes")
        return torch.mean(loss)



from ptls.data_load.iterable_processing.target_extractor import TargetExtractor


def get_features_only():
    return [
        {'id': 1, 'target': 0},
        {'id': 2, 'target': 0},
        {'id': 3, 'target': 1},
        {'id': 4, 'target': 1},
    ]


def get_features_with_incorrect_target():
    return [
        ({'id': 1, 'target': 0}, 0),
        ({'id': 2, 'target': 0}, 1),
        ({'id': 3, 'target': 1}, 0),
        ({'id': 4, 'target': 1}, 1),
    ]


def test_extract():
    i_filter = TargetExtractor('target')
    data = i_filter(get_features_only())
    data = [y for x, y in data]
    assert data == [0, 0, 1, 1]


def test_extract_and_replace():
    i_filter = TargetExtractor('target')
    orig_data = get_features_with_incorrect_target()

    data = orig_data
    data = [y for x, y in data]
    assert data == [0, 1, 0, 1]  # without replace

    data = i_filter(orig_data)
    data = [y for x, y in data]
    assert data == [0, 0, 1, 1]  # with replace





"""
Creates data loaders for Next Sequence Prediction learning
Each record pass throw the sampler and splits into views of the main record
Each sample slits into left and right part.
Some right part are shuffled.
X - is pair of feature_arrays (left and right), sub-samples of main event sequences
Y - is left and right from she same sample.
    0 means left and right from different sample, 1 means left and right from the same sample

    Input:        -> Output
dict of arrays    ->  X,                                            Y
client_1          -> (client_1_smpl_1_left, client_1_smpl_1_right), 1
                     (client_1_smpl_2_left, client_1_smpl_2_right), 1
                     (client_1_smpl_3_left, client_3_smpl_1_right), 0
client_2          -> (client_2_smpl_1_left, client_2_smpl_3_right), 0
                     ...
"""
import logging
import random
import warnings

import pytorch_lightning as pl
import torch

from torch.utils.data import DataLoader
from tqdm.auto import tqdm

from ptls.data_load.augmentations.build_augmentations import build_augmentations
from ptls.data_load.augmentations.sequence_pair_augmentation import sequence_pair_augmentation
from ptls.data_load import IterableChain, augmentation_chain, padded_collate_wo_target
from ptls.data_load.iterable_processing.category_size_clip import CategorySizeClip
from ptls.data_load.iterable_processing.feature_filter import FeatureFilter
from ptls.data_load.iterable_processing.iterable_shuffle import IterableShuffle
from ptls.data_load.iterable_processing.seq_len_filter import SeqLenFilter
from ptls.data_load.list_splitter import ListSplitter
from ptls.data_load.datasets.parquet_dataset import ParquetDataset, ParquetFiles
from ptls.data_load.partitioned_dataset import PartitionedDataset, PartitionedDataFiles
from ptls.metric_learn.dataset import nested_list_to_flat_with_collate
from ptls.frames.coles import split_strategy
from ptls.metric_learn.dataset.splitting_dataset import IterableSplittingDataset, MapSplittingDataset

logger = logging.getLogger(__name__)


def collate_nsp_pairs(batch):
    lefts = [left for left, _ in batch] * 2

    rights = [right for _, right in batch]
    rights_ = rights[:]
    random.shuffle(rights_)
    rights += rights_

    targets = torch.cat([
        torch.ones(len(batch), dtype=torch.int64),
        torch.zeros(len(batch), dtype=torch.int64),
    ])

    return (
        (
            padded_collate_wo_target(lefts),
            padded_collate_wo_target(rights)
        ),
        targets.long(),
    )


class NspDataModuleTrain(pl.LightningDataModule):
    def __init__(self, type, setup, train, valid, pl_module):
        warnings.warn('Use `ptls.frames.PtlsDataModule` '
                      'with `ptls.frames.bert.NspDataset` or `ptls.frames.bert.NspIterableDataset`',
                      DeprecationWarning)
        super().__init__()

        self._type = type
        assert self._type in ('map', 'iterable')

        self.setup_conf = setup
        self.train_conf = train
        self.valid_conf = valid

        self.col_id = self.setup_conf.col_id
        self.category_names = pl_module.seq_encoder.category_names
        self.category_names.add('event_time')
        self.category_max_size = pl_module.seq_encoder.category_max_size

        self.train_dataset = None
        self.valid_dataset = None
        self._train_ids = None
        self._valid_ids = None

    def prepare_data(self):
        if 'dataset_files' in self.setup_conf:
            self.setup_iterable_files()
        elif 'dataset_parts' in self.setup_conf:
            self.setup_iterable_parts()
        else:
            raise AttributeError(f'Missing dataset definition. One of `dataset_parts` or `dataset_files` expected')

        if self._type == 'map':
            self.setup_map()

    def setup_iterable_files(self):
        if self.setup_conf.split_by == 'files':
            data_files = ParquetFiles(self.setup_conf.dataset_files.data_path).data_files

            splitter = ListSplitter(
                data_files,
                valid_size=self.setup_conf.valid_size,
                seed=self.setup_conf.valid_split_seed,
            )
            logger.info(f'Prepared splits: '
                        f'{len(splitter.train)} files in train, {len(splitter.valid)} files in valid')

            self.train_dataset = ParquetDataset(
                splitter.train,
                post_processing=IterableChain(*self.build_iterable_processing('train')),
                shuffle_files=True if self._type == 'iterable' else False,
            )
            self.valid_dataset = ParquetDataset(
                splitter.valid,
                post_processing=IterableChain(*self.build_iterable_processing('valid')),
                shuffle_files=False,
            )

        else:
            raise AttributeError(f'Unknown split strategy: {self.setup_conf.split_by}')

    def setup_iterable_parts(self):
        data_files = PartitionedDataFiles(**self.setup_conf.dataset_parts)

        if self.setup_conf.split_by == 'hash_id':
            splitter = ListSplitter(
                data_files.hash_parts,
                valid_size=self.setup_conf.valid_size,
                seed=self.setup_conf.valid_split_seed,
            )
            logger.info(f'Prepared splits: '
                        f'{len(data_files.dt_parts)} parts in dt_train, {len(data_files.dt_parts)} parts in dt_valid, '
                        f'{len(splitter.train)} parts in hash_train, {len(splitter.valid)} parts in hash_valid')

            self.train_dataset = PartitionedDataset(
                data_files.data_path, data_files.dt_parts, splitter.train, col_id=self.col_id,
                post_processing=IterableChain(*self.build_iterable_processing('train')),
                shuffle_files=True if self._type == 'iterable' else False,
            )
            self.valid_dataset = PartitionedDataset(
                data_files.data_path, data_files.dt_parts, splitter.valid, col_id=self.col_id,
                post_processing=IterableChain(*self.build_iterable_processing('valid')),
                shuffle_files=False,
            )
        elif self.setup_conf.split_by == 'rows':
            raise NotImplementedError("Split by rows aren't supported for partitioned dataset")
        else:
            raise AttributeError(f'Unknown split strategy: {self.setup_conf.split_by}')

    def build_iterable_processing(self, part):
        if part == 'train':
            yield SeqLenFilter(min_seq_len=self.train_conf.min_seq_len)

        yield FeatureFilter(keep_feature_names=self.category_names)
        yield CategorySizeClip(self.category_max_size)

        if self._type == 'iterable':
            # all processing in single chain
            if part == 'train':
                yield IterableShuffle(self.train_conf.buffer_size)

            if part == 'train':
                split_strategy_conf = self.train_conf.split_strategy
            else:
                split_strategy_conf = self.valid_conf.split_strategy
            yield IterableSplittingDataset(split_strategy.create(**split_strategy_conf),
                                           self.build_augmentations(part))

    def build_augmentations(self, part):
        if part == 'train':
            a_chain = build_augmentations(self.train_conf.augmentations)
        else:
            a_chain = build_augmentations(self.valid_conf.augmentations)
        return augmentation_chain(a_chain, sequence_pair_augmentation)

    def setup_map(self):
        self.train_dataset = list(tqdm(iter(self.train_dataset)))
        logger.info(f'Loaded {len(self.train_dataset)} for train')
        self.valid_dataset = list(tqdm(iter(self.valid_dataset)))
        logger.info(f'Loaded {len(self.valid_dataset)} for valid')

        self.train_dataset = MapSplittingDataset(
            base_dataset=self.train_dataset,
            splitter=split_strategy.create(**self.train_conf.split_strategy),
            a_chain=self.build_augmentations('train'),
        )
        self.valid_dataset = MapSplittingDataset(
            base_dataset=self.valid_dataset,
            splitter=split_strategy.create(**self.valid_conf.split_strategy),
            a_chain=self.build_augmentations('valid'),
        )

    def train_dataloader(self):
        return DataLoader(
            dataset=self.train_dataset,
            collate_fn=nested_list_to_flat_with_collate(collate_nsp_pairs),
            shuffle=False if self._type == 'iterable' else True,
            num_workers=self.train_conf.num_workers,
            batch_size=self.train_conf.batch_size,
        )

    def val_dataloader(self):
        return DataLoader(
            dataset=self.valid_dataset,
            collate_fn=nested_list_to_flat_with_collate(collate_nsp_pairs),
            num_workers=self.valid_conf.num_workers,
            batch_size=self.valid_conf.batch_size,
        )



import pytorch_lightning as pl
import torch
from torch import nn
import warnings
from torchmetrics import MeanMetric
from typing import Tuple, Dict

from ptls.nn.seq_encoder.abs_seq_encoder import AbsSeqEncoder
from ptls.nn import PBL2Norm
from ptls.data_load.padded_batch import PaddedBatch
from ptls.custom_layers import StatPooling


class TabformerPretrainModule(pl.LightningModule):
    """Tabformer Model (MLM) from [Tabular Transformers for Modeling Multivariate Time Series](https://arxiv.org/abs/2011.01843)

    Original sequence are encoded by `TrxEncoder`.
    Randomly sampled features are replaced by MASK token.
    Transformer `seq_encoder` reconstruct masked tokens.
    The loss function is a classification loss.

    Parameters
    ----------
    trx_encoder:
        Module for transform dict with feature sequences to sequence of transaction representations
    feature_encoder:
        Module that apply transformer layers to output of trx_encoder on feature dimension, (B, T, H_in) -> (B, T, H_out)
    seq_encoder:
        Module for sequence processing. Generally this is transformer based encoder. Rnn is also possible
        Should works without sequence reduce
    total_steps:
        total_steps expected in OneCycle lr scheduler
    max_lr:
        max_lr of OneCycle lr scheduler
    weight_decay:
        weight_decay of Adam optimizer
    pct_start:
        % of total_steps when lr increase
    norm_predict:
        use l2 norm for transformer output or not
    mask_prob:
        probability of masking randomly selected feature
    inference_pooling_strategy:
        'out' - `seq_encoder` forward (`is_reduce_requence=True`) (B, H)
        'stat' - min, max, mean, std statistics pooled from `trx_encoder` layer + 'out' from `seq_encoder` (B, H) -> (B, 5H)
    """

    def __init__(self,
                 trx_encoder: torch.nn.Module,
                 feature_encoder: torch.nn.Module,
                 seq_encoder: AbsSeqEncoder,
                 total_steps: int,
                 max_lr: float = 0.001,
                 weight_decay: float = 0.0,
                 pct_start: float = 0.1,
                 norm_predict: bool = False,
                 mask_prob: float = 0.15,
                 inference_pooling_strategy: str = 'out'
                 ):

        super().__init__()
        self.save_hyperparameters(ignore=['trx_encoder', 'seq_encoder', 'feature_encoder'])

        self.trx_encoder = trx_encoder
        self.feature_encoder = feature_encoder

        assert not self.trx_encoder.numeric_values, '`numeric_values` parameter of `trx_encoder` should be == {}. Discretize all numerical features into categorical to use Tabformer model!'
        noisy_embeds = list(self.trx_encoder.embeddings.values())
        assert noisy_embeds, '`embeddings` parameter for `trx_encoder` should contain at least 1 feature!'
        self.feature_emb_dim = noisy_embeds[0].embedding_dim
        self.num_f = len(self.trx_encoder.embeddings)

        self.head = nn.ModuleList()
        in_head_dim = self.feature_emb_dim * self.num_f
        for n_emb in noisy_embeds:
            self.head += [nn.Linear(in_head_dim, n_emb.num_embeddings+1)]
            assert all(n_emb.embedding_dim == self.feature_emb_dim for n_emb in noisy_embeds), 'Out dimensions for all features in `embeddings` parameter of `trx_encoder` should be equal for Tabformer model!'

        warnings.warn("With Tabformer model set `in` value in `embeddings`parameter of `trx_encoder` equal to actual number of unique feature values + 1")

        self._seq_encoder = seq_encoder
        self._seq_encoder.is_reduce_sequence = False

        if self.hparams.norm_predict:
            self.fn_norm_predict = PBL2Norm()

        self.token_mask = torch.nn.Parameter(torch.randn(1, 1, self.feature_emb_dim), requires_grad=True)

        self.loss = nn.CrossEntropyLoss()

        self.train_tabformer_loss = MeanMetric()
        self.valid_tabformer_loss = MeanMetric()
        self.mask_prob = mask_prob

        self.lin_proj = nn.Sequential(nn.Linear(self.feature_emb_dim, self.feature_emb_dim * self.num_f),
                                      nn.GELU(),
                                      nn.LayerNorm(self.feature_emb_dim * self.num_f, eps=1e-12)
                                      )

    def configure_optimizers(self):
        optim = torch.optim.Adam(self.parameters(),
                                 lr=self.hparams.max_lr,
                                 weight_decay=self.hparams.weight_decay,
                                 )
        scheduler = torch.optim.lr_scheduler.OneCycleLR(
            optimizer=optim,
            max_lr=self.hparams.max_lr,
            total_steps=self.hparams.total_steps,
            pct_start=self.hparams.pct_start,
            anneal_strategy='cos',
            cycle_momentum=False,
            div_factor=25.0,
            final_div_factor=10000.0,
            three_phase=False,
        )
        scheduler = {'scheduler': scheduler, 'interval': 'step'}
        return [optim], [scheduler]

    def forward(self, z: PaddedBatch):
        out = self._seq_encoder(z)
        if self.hparams.norm_predict:
            out = self.fn_norm_predict(out)
        return out

    def get_masks_and_labels(self, batch: PaddedBatch) -> (torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor):
        feature_tensors, random_words = [], []
        for field_name, noisy_emb_module in self.trx_encoder.embeddings.items():
            feature_tensors += [batch.payload[field_name].unsqueeze(0)]
            random_words += [noisy_emb_module(torch.randint(1, noisy_emb_module.num_embeddings, batch.seq_len_mask.shape, dtype=torch.long).to(batch.device)).unsqueeze(0)]
        feature_tensors = torch.cat(feature_tensors)
        random_words = torch.cat(random_words).permute(1, 2, 0, 3)
        pad_mask_tensors = batch.seq_len_mask.unsqueeze(0).repeat((feature_tensors.shape[0], 1, 1))

        pad_tokens_mask = ~pad_mask_tensors.type(torch.bool)
        probability_matrix = torch.full(pad_tokens_mask.shape, self.mask_prob).to(batch.device)
        probability_matrix.masked_fill_(pad_tokens_mask, value=0.0)

        return tuple(*self.tabformer_mask(feature_tensors, probability_matrix), random_words)

    def tabformer_mask(self, inputs: torch.Tensor, probability_matrix: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
            """
            Prepare masked tokens inputs/labels for masked language modeling: 80% MASK, 10% random, 10% original.
            """
            labels = inputs.clone()
            # We sample a few tokens in each sequence for masked-LM training (with probability defined in probability_matrix
            # defaults to 0.15 in Bert/RoBERTa)
            masked_indices = torch.bernoulli(probability_matrix).bool().to(inputs.device)
            labels[~masked_indices] = -100  # We only compute loss on masked tokens
    
            # 80% of the time, we replace masked input tokens with [MASK] token
            indices_replaced = (torch.bernoulli(torch.full(labels.shape, 0.8)).bool().to(inputs.device) & masked_indices)
    
            # 10% of the time, we replace masked input tokens with random word
            # The rest of the time (10% of the time) we keep the masked input tokens unchanged
            indices_random = (torch.bernoulli(torch.full(labels.shape, 0.5)).bool().to(inputs.device) & masked_indices & ~indices_replaced)
    
            return labels.permute(1, 2, 0), masked_indices.permute(1, 2, 0), indices_replaced.permute(1, 2, 0)

    def loss_tabformer(self, x: PaddedBatch, target, is_train_step):
        out = self.forward(x)
        sequence_output = out.payload
        masked_lm_labels = target.view(-1, self.num_f).permute(1, 0)  # (B, T, NUM_F) -> (NUM_F, B*T)

        expected_sz = (-1, self.num_f, self.feature_emb_dim)
        sequence_output = sequence_output.reshape(expected_sz).permute(1, 0, 2)  # (B, T, NUM_F*H) -> (NUM_F, B*T, H)
        sequence_output = self.lin_proj(sequence_output)

        seq_out_feature = torch.chunk(sequence_output, self.num_f, dim=0)
        labels_feature = torch.chunk(masked_lm_labels, self.num_f, dim=0)

        loss = 0
        for f_ix in range(self.num_f):
            seq_out_f, labels_f = seq_out_feature[f_ix].squeeze(0), labels_feature[f_ix].squeeze(0)
            out_f = self.head[f_ix](seq_out_f)
            loss_f = self.loss(out_f, labels_f)
            loss += loss_f
        return loss

    def training_step(self, batch, batch_idx):
        tabf_labels, MASK_token_mask, RANDOM_token_mask, random_words = self.get_masks_and_labels(batch)
        z_trx = self.trx_encoder(batch)  # PB: B, T, H
        
        payload = z_trx.payload.view(z_trx.payload.shape[:-1] + (-1, self.feature_emb_dim))
        payload[MASK_token_mask] = self.token_mask
        payload[RANDOM_token_mask] = random_words[RANDOM_token_mask]
        payload = self.feature_encoder(payload)

        z_trx._payload = payload

        loss_tabformer = self.loss_tabformer(z_trx, tabf_labels, is_train_step=True)
        self.train_tabformer_loss(loss_tabformer)
        self.log(f'tabformer/loss', loss_tabformer)
        return loss_tabformer

    def validation_step(self, batch, batch_idx):
        tabf_labels, MASK_token_mask, RANDOM_token_mask, random_words = self.get_masks_and_labels(batch)
        z_trx = self.trx_encoder(batch)  # PB: B, T, H

        payload = z_trx.payload.view(z_trx.payload.shape[:-1] + (-1, self.feature_emb_dim))
        payload[MASK_token_mask] = self.token_mask
        payload[RANDOM_token_mask] = random_words[RANDOM_token_mask]
        payload = self.feature_encoder(payload)

        z_trx._payload = payload

        loss_tabformer = self.loss_tabformer(z_trx, tabf_labels, is_train_step=False)
        self.valid_tabformer_loss(loss_tabformer)

    def training_epoch_end(self, _):
        self.log(f'tabformer/train_tabformer_loss', self.train_tabformer_loss, prog_bar=False)
        # self.train_tabformer_loss reset not required here

    def validation_epoch_end(self, _):
        self.log(f'tabformer/valid_tabformer_loss', self.valid_tabformer_loss, prog_bar=True)
        # self.valid_tabformer_loss reset not required here

    @property
    def seq_encoder(self):
        return TabformerInferenceModule(pretrained_model=self)


class TabformerInferenceModule(torch.nn.Module):
    def __init__(self, pretrained_model):
        super().__init__()
        self.model = pretrained_model
        self.model._seq_encoder.is_reduce_sequence = True
        self.model._seq_encoder.add_cls_output = False 
        if self.model.hparams.inference_pooling_strategy=='stat':
            self.stat_pooler = StatPooling()

    def forward(self, batch: PaddedBatch):
        z_trx = self.model.trx_encoder(batch)
        payload = z_trx.payload.view(z_trx.payload.shape[:-1] + (-1, self.model.feature_emb_dim))
        payload = self.model.feature_encoder(payload)
        encoded_trx = PaddedBatch(payload=payload, length=z_trx.seq_lens)
        out = self.model._seq_encoder(encoded_trx)

        if self.model.hparams.inference_pooling_strategy=='stat':
            stats = self.stat_pooler(z_trx)
            out = torch.cat([stats, out], dim=1)  # out: B, 5H
        if self.model.hparams.norm_predict:
            out = out / (out.pow(2).sum(dim=-1, keepdim=True) + 1e-9).pow(0.5)
        return out


import logging
import random
import math
import torch
from torch import nn

from ptls.data_load.padded_batch import PaddedBatch
from ptls.nn.seq_encoder.abs_seq_encoder import AbsSeqEncoder

logger = logging.getLogger(__name__)


class PositionalEncoding(nn.Module):
    def __init__(self,
                 d_model,
                 use_start_random_shift=True,
                 max_len=5000,
                 ):
        super(PositionalEncoding, self).__init__()
        self.use_start_random_shift = use_start_random_shift
        self.max_len = max_len

        pe = torch.zeros(max_len, d_model)
        position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2).float() * (-math.log(10000.0) / d_model))
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0)
        self.register_buffer('pe', pe)

    def forward(self, x):
        T = x.size(1)
        if self.training and self.use_start_random_shift:
            start_pos = random.randint(0, self.max_len - T)
        else:
            start_pos = 0
        x = x + self.pe[:, start_pos:start_pos + T]
        return x


class TransformerEncoder(AbsSeqEncoder):
    """Used torch implementation of transformer
    Based on `torch.nn.TransformerEncoder`

    Parameters
        input_size:
            input embedding size.
            Equals intermediate and output layer size cause transformer don't change vector dimentions
        train_starter:
            'randn' or 'zeros'
            Which token used for CLS token, random learnable or zeros fixed
        shared_layers:
            True - then the same weights used for all `n_layers`.
            False - `n_layers` used different weights
        n_heads:
            The number of heads in the multiheadattention models
        dim_hidden:
            The dimension of the feedforward network model
        dropout:
            The dropout value
        n_layers:
            The number of sub-encoder-layers in the encoder
        use_positional_encoding (bool):
            Use or not positional encoding
        use_start_random_shift (bool):
            True - starting pos of positional encoding randomly shifted when training
            This allow to train transformer with all range of positional encoding values
            False - starting pos is not shifted.
        max_seq_len:
            The possible maximum sequence length for positional encoding
        use_after_mask:
            True value makes transformer unidirectional
        use_src_key_padding_mask:
            Padding simbols aren't used in attention bases on sequences lenghts
        use_norm_layer:
            Use or not LayerNorm
        is_reduce_sequence (bool):
            False - returns PaddedBatch with all transactions embeddings
            True - returns one embedding for sequence based on CLS token

    Example:
    >>> model = TransformerEncoder(input_size=32)
    >>> x = PaddedBatch(torch.randn(10, 128, 32), torch.randint(20, 128, (10,)))
    >>> y = model(x)
    >>> assert y.payload.size() == (10, 128, 32)
    >>>
    >>> model = TransformerEncoder(input_size=32, is_reduce_sequence=True)
    >>> y = model(x)
    >>> assert y.size() == (10, 32)

    """
    def __init__(self,
                 input_size,
                 starter='randn',
                 shared_layers=False,
                 n_heads=8,
                 dim_hidden=256,
                 dropout=0.1,
                 n_layers=6,
                 use_positional_encoding=True,
                 use_start_random_shift=True,
                 max_seq_len=5000,
                 use_after_mask=False,
                 use_src_key_padding_mask=True,
                 use_norm_layer=True,
                 is_reduce_sequence=False,  # previous default behavior for TransformerSeqEncoder
                 ):
        super().__init__(is_reduce_sequence=is_reduce_sequence)

        self.input_size = input_size
        self.shared_layers = shared_layers
        self.n_layers = n_layers
        self.use_after_mask = use_after_mask
        self.use_src_key_padding_mask = use_src_key_padding_mask
        self.use_positional_encoding = use_positional_encoding

        if starter == 'randn':
            self.starter = torch.nn.Parameter(torch.randn(1, 1, input_size), requires_grad=True)
        elif starter == 'zeros':
            self.starter = torch.nn.Parameter(torch.zeros(1, 1, input_size), requires_grad=False)
        else:
            raise AttributeError(f'Unknown train_starter: "{starter}". Expected one of [randn, zeros]')

        enc_layer = torch.nn.TransformerEncoderLayer(
            d_model=input_size,
            nhead=n_heads,
            dim_feedforward=dim_hidden,
            dropout=dropout,
            batch_first=True,
        )
        enc_norm = torch.nn.LayerNorm(input_size) if use_norm_layer else None

        if self.shared_layers:
            self.enc_layer = enc_layer
            self.enc_norm = enc_norm
        else:
            self.enc = torch.nn.TransformerEncoder(enc_layer, n_layers, enc_norm)

        if self.use_positional_encoding:
            self.pe = PositionalEncoding(
                use_start_random_shift=use_start_random_shift,
                max_len=max_seq_len,
                d_model=input_size,
            )

    @staticmethod
    def generate_square_subsequent_mask(sz):
        r"""Generate a square mask for the sequence. The masked positions are filled with float('-inf').
            Unmasked positions are filled with float(0.0).
        """
        mask = (torch.triu(torch.ones(sz, sz)) == 1).transpose(0, 1)
        mask = mask.float().masked_fill(mask == 0, float('-inf')).masked_fill(mask == 1, float(0.0))
        mask[0, :] = 0.0
        mask[:, 0] = 0.0
        return mask

    def forward(self, x: PaddedBatch):
        B, T, H = x.payload.size()

        if self.use_after_mask:
            src_mask = self.generate_square_subsequent_mask(T + 1).to(x.device)
        else:
            src_mask = None

        if self.use_src_key_padding_mask:
            src_key_padding_mask = torch.cat([
                torch.zeros(B, 1, dtype=torch.long, device=x.device),
                (1 - x.seq_len_mask),
            ], dim=1).bool()
        else:
            src_key_padding_mask = None

        x_in = x.payload
        if self.use_positional_encoding:
            x_in = self.pe(x_in)
        x_in = torch.cat([self.starter.expand(B, 1, H), x_in], dim=1)

        if self.shared_layers:
            out = x_in
            for _ in range(self.n_layers):
                out = self.enc_layer(out, src_mask=src_mask, src_key_padding_mask=src_key_padding_mask)
                if self.enc_norm is not None:
                    out = self.enc_norm(out)
        else:
            out = self.enc(x_in, mask=src_mask, src_key_padding_mask=src_key_padding_mask)

        if self.is_reduce_sequence:
            return out[:, 0, :]

        return PaddedBatch(out[:, 1:, :], x.seq_lens)

    @property
    def embedding_size(self):
        return self.input_size



import torch

from ptls.nn.trx_encoder.noisy_embedding import NoisyEmbedding


def test_no_noise():
    embedding = NoisyEmbedding(16, 4, 0, noise_scale=0.0)
    x = torch.zeros(4, 8, dtype=torch.long)
    out = embedding(x)
    assert out.size() == (4, 8, 4)


def test_noisy_embedding():
    embedding = NoisyEmbedding(16, 4, 0, noise_scale=1.0)
    x = torch.zeros(4, 8, dtype=torch.long)
    out = embedding(x)
    assert out.size() == (4, 8, 4)


def test_spatial_dropout():
    embedding = NoisyEmbedding(16, 4, 0, noise_scale=0.0, dropout=0.27, spatial_dropout=True)
    x = torch.ones(4000, 8000, dtype=torch.long)
    out = embedding(x)
    nonzero = torch.count_nonzero(out, dim=1)
    assert torch.all((nonzero == 8000) + (nonzero == 0)).item()
    assert torch.abs(0.27 - (1 - torch.count_nonzero(out) / 4000 / 8000 / 4)) < 0.01


def test_dropout():
    embedding = NoisyEmbedding(16, 4, 0, noise_scale=0.0, dropout=0.35, spatial_dropout=False)
    x = torch.ones(400, 800, dtype=torch.long)
    out = embedding(x)
    assert torch.abs(0.35 - (1 - torch.count_nonzero(out) / 400 / 800 / 4)) < 0.01



import random
import torch
from transformers import LongformerConfig, LongformerModel

from ptls.data_load.padded_batch import PaddedBatch
from ptls.nn.seq_encoder.abs_seq_encoder import AbsSeqEncoder


class LongformerEncoder(AbsSeqEncoder):
    """Used huggingface implementation of transformer
    Based on `transformers.LongformerModel`
    [Link](https://huggingface.co/docs/transformers/main/en/model_doc/longformer#transformers.LongformerModel)

    Transformer-based models are unable to process long sequences due to their self-attention operation,
    which scales quadratically with the sequence length. To address this limitation, was introduce the Longformer
    with an attention mechanism that scales linearly with sequence length, making it easy to process documents
    of thousands of tokens or longer. Longformer’s attention mechanism is a drop-in replacement for
    the standard self-attention and combines a local windowed attention with a task motivated global attention.

    Parameters
        input_size:
            input embedding size.
            Equals intermediate and output layer size cause transformer don't change vector dimentions
        num_attention_heads:
            The number of heads in the multiheadattention models
        intermediate_size:
            The dimension of the feedforward network model
        num_hidden_layers:
            The number of sub-encoder-layers in the encoder
        attention_window:
            Size of an attention window around each token
        max_position_embeddings:
            The possible maximum sequence length for positional encoding
        use_positional_encoding (bool):
            Use or not positional encoding
        use_start_random_shift (bool):
            True - starting pos of positional encoding randomly shifted when training
            This allow to train transformer with all range of positional encoding values
            False - starting pos is not shifted.
        is_reduce_sequence (bool):
            False - returns PaddedBatch with all transactions embeddings
            True - returns one embedding for sequence based on CLS token
        add_cls_output (bool):
            False - returns PaddedBatch with all transactions embeddings
            True - returns tuple (PaddedBatch with all transactions embeddings, one embedding for sequence based on CLS token)
    Example:
    >>> model = LongformerEncoder(input_size=32)
    >>> x = PaddedBatch(torch.randn(10, 128, 32), torch.randint(20, 128, (10,)))
    >>> y = model(x)
    >>> assert y.payload.size() == (10, 128, 32)
    >>>
    >>> model = LongformerEncoder(input_size=32, is_reduce_sequence=True)
    >>> y = model(x)
    >>> assert y.size() == (10, 32)

    """
    def __init__(self,
                 input_size,
                 num_attention_heads: int = 1,
                 intermediate_size: int = 128,
                 num_hidden_layers: int = 1,
                 attention_window: int = 16,
                 hidden_act='gelu',
                 max_position_embeddings=5000,
                 use_positional_encoding=True,
                 use_start_random_shift=True,
                 is_reduce_sequence=False,
                 add_cls_output=False,
                 ):
        super().__init__(is_reduce_sequence=is_reduce_sequence)

        self.hidden_size = input_size
        self.max_position_embeddings = max_position_embeddings
        self.use_positional_encoding = use_positional_encoding
        self.use_start_random_shift = use_start_random_shift

        self.token_cls = torch.nn.Parameter(torch.randn(1, 1, input_size), requires_grad=True)
        self.transf = LongformerModel(
            config=LongformerConfig(
                hidden_size=input_size,
                num_attention_heads=num_attention_heads,
                intermediate_size=intermediate_size,
                num_hidden_layers=num_hidden_layers,
                hidden_act=hidden_act,
                vocab_size=4,
                max_position_embeddings=max_position_embeddings,
                attention_window=attention_window,
            ),
            add_pooling_layer=False,
        )
        self.add_cls_output=add_cls_output

    def forward(self, x: PaddedBatch):
        B, T, H = x.payload.size()
        device = x.device

        if self.training and self.use_start_random_shift:
            start_pos = random.randint(0, self.max_position_embeddings - T - 1)
        else:
            start_pos = 0

        inputs_embeds = torch.cat([
            self.token_cls.expand(B, 1, H),
            x.payload,
        ], dim=1)
        attention_mask = torch.cat([
            torch.ones(B, 1, device=device),
            x.seq_len_mask.float(),
        ], dim=1)
        if self.use_positional_encoding:
            position_ids = torch.arange(T, device=device, dtype=torch.long).view(1, -1).expand(B, T) + 1 + start_pos
        else:
            position_ids = torch.zeros(B, T, device=device, dtype=torch.long)
        position_ids = torch.cat([
            torch.zeros(B, 1, device=device, dtype=torch.long),
            position_ids,
        ], dim=1)
        global_attention_mask = torch.cat([
            torch.ones(B, 1, device=device, dtype=torch.long),
            torch.zeros(B, T, device=device, dtype=torch.long),
        ], dim=1)

        out = self.transf(
            inputs_embeds=inputs_embeds,
            attention_mask=attention_mask,
            position_ids=position_ids,
            global_attention_mask=global_attention_mask,
        ).last_hidden_state

        if self.is_reduce_sequence:
            return out[:, 0, :]
        else:
            if self.add_cls_output:
                return PaddedBatch(out[:, 1:, :], x.seq_lens), out[:, 0, :]
            else:
                return PaddedBatch(out[:, 1:, :], x.seq_lens)

    @property
    def embedding_size(self):
        return self.hidden_size



from typing import List, Dict

from torch.utils.data import DataLoader

from ptls.data_load import IterableChain, padded_collate_wo_target
from ptls.data_load.filter_dataset import FilterDataset
from ptls.data_load.iterable_processing import ToTorch, FilterNonArray, ISeqLenLimit


def inference_data_loader(
    data: List[Dict],
    max_seq_len: int = 10000,
    num_workers: int = 0,
    batch_size: int = 512,
):
    r"""Generate an inference data loader

    Parameters
    ----------
    data: List[Dict]
        The dataset
    max_seq_len:
        Only `max_seq_len` transaction will taken from tail of sequence.
        Unlimited sequence length lead to out of memory error
    num_workers: int. Default: 0.
        The number of workers for the dataloader. 0 = single-process loader
    batch_size: int. Default: 512.
        The number of samples (before splitting to subsequences) in each batch
    """
    dataset = FilterDataset(
        data,
        post_processing=IterableChain(
            ToTorch(),
            FilterNonArray(),
            ISeqLenLimit(max_seq_len=max_seq_len),
        )
    )

    return DataLoader(
        dataset=dataset,
        collate_fn=padded_collate_wo_target,
        shuffle=False,
        num_workers=num_workers,
        batch_size=batch_size,
    )



from ptls.data_load.iterable_processing.seq_len_filter import SeqLenFilter
import numpy as np


def get_data(id_type=int, array_type=np.array):
    return [
        {
            'uid': id_type(1),
            'seq_len': 10,
            'mcc_code': array_type([1, 2, 3, 4]),
            'amount': array_type([1, 2, 3, 4]),
        },
        {
            'uid': id_type(2),
            'seq_len': 11,
            'mcc_code': array_type([1, 2, 3, 4, 5, 6, 7]),
            'amount': array_type([1, 2, 3, 4, 5, 6, 7]),
        },
        {
            'uid': id_type(3),
            'seq_len': 12,
            'mcc_code': array_type([1, 2, 3, 4, 5, 6, 7, 8, 9]),
            'amount': array_type([1, 2, 3, 4, 6, 7, 8, 9]),
        },
    ]


def get_data_with_target():
    return [
        ({
            'uid': "1",
            'mcc_code': np.array([1, 2, 3, 4]),
            'amount': np.array([1, 2, 3, 4],),
            'target': 1,
        }, 1),
        ({
             'uid': "2",
             'mcc_code': np.array([1, 2]),
             'amount': np.array([1, 2]),
             'target': 0,
         }, 0),
    ]


def test_no_action():
    i_filter = SeqLenFilter(sequence_col='mcc_code')
    data = i_filter(get_data())
    data = [rec['uid'] for rec in data]
    assert data == [1, 2, 3]


def test_min_len():
    i_filter = SeqLenFilter(sequence_col='mcc_code', min_seq_len=5)
    data = i_filter(get_data())
    data = [rec['uid'] for rec in data]
    assert data == [2, 3]


def test_max_len():
    i_filter = SeqLenFilter(sequence_col='mcc_code', max_seq_len=8)
    data = i_filter(get_data())
    data = [rec['uid'] for rec in data]
    assert data == [1, 2]


def test_seq_len_field():
    i_filter = SeqLenFilter(seq_len_col='seq_len', min_seq_len=12)
    data = i_filter(get_data())
    data = [rec['uid'] for rec in data]
    assert data == [3]


def test_target_col_detection():
    i_filter = SeqLenFilter(min_seq_len=5, max_seq_len=8)
    data = i_filter(get_data())
    data = [rec['uid'] for rec in data]
    assert data == [2]


def test_target_col_detection_str():
    i_filter = SeqLenFilter(min_seq_len=5, max_seq_len=8)
    data = i_filter(get_data(id_type=str))
    data = [rec['uid'] for rec in data]
    assert data == ['2']


def test_with_target():
    i_filter = SeqLenFilter(min_seq_len=3)
    data = i_filter(get_data_with_target())
    data = [rec[0]['uid'] for rec in data]
    assert data == ['1']



import torch
from torch import nn as nn
from torch.nn import functional as F


class ContrastiveLoss(nn.Module):
    """
    Contrastive loss

    "Signature verification using a siamese time delay neural network", NIPS 1993
    https://papers.nips.cc/paper/769-signature-verification-using-a-siamese-time-delay-neural-network.pdf
    """

    def __init__(self, margin, sampling_strategy):
        super(ContrastiveLoss, self).__init__()
        self.margin = margin
        self.pair_selector = sampling_strategy

    def forward(self, embeddings, target):

        positive_pairs, negative_pairs = self.pair_selector.get_pairs(embeddings, target)
        positive_loss = F.pairwise_distance(embeddings[positive_pairs[:, 0]], embeddings[positive_pairs[:, 1]]).pow(2)

        negative_loss = F.relu(
            self.margin - F.pairwise_distance(embeddings[negative_pairs[:, 0]], embeddings[negative_pairs[:, 1]])
        ).pow(2)
        loss = torch.cat([positive_loss, negative_loss], dim=0)

        return loss.sum()


import numpy as np

from ptls.data_load.iterable_processing_dataset import IterableProcessingDataset


class TargetEmptyFilter(IterableProcessingDataset):
    def __init__(self, target_col):
        """Drop records where value in `target_col` is undefined

        Args:
            target_col: field where `y` is stored
        """
        super().__init__()

        self._target_col = target_col

    def __iter__(self):
        for rec in self._src:
            features = rec[0] if type(rec) is tuple else rec
            y = features[self._target_col]
            if y is None:
                continue
            if type(y) is not str and np.isnan(y):
                continue
            yield rec



import torch

from ptls.frames.bert.datasets.mlm_indexed_dataset import MlmIndexedDataset


def get_data():
    return [
        {
            'cat': torch.Tensor([1, 2, 3, 4, 5, 6]),
            'amnt': torch.Tensor([10, 12, 13, 41, 24]),
        },
        {
            'cat': torch.Tensor([9, 8, 7, 6, 5, 4, 3, 2]),
            'amnt': torch.Tensor([11, 16, 17, 16, 13, 10, 18, 16]),
        },
    ]


def test_mlm_dataset():
    ds = MlmIndexedDataset(
        data=get_data(),
        seq_len=3,
    )
    assert len(ds) == 5

    item0 = ds[0]
    assert item0['cat'].tolist() == [1, 2, 3]
    assert item0['amnt'].tolist() == [10, 12, 13]

    item4 = ds[4]
    assert item4['cat'].tolist() == [4, 3, 2]
    assert item4['amnt'].tolist() == [10, 18, 16]


def test_mlm_dataloader():
    ds = MlmIndexedDataset(
        data=get_data(),
        seq_len=3,
        step_rate=2/3,
    )
    dl = torch.utils.data.DataLoader(ds, batch_size=3, collate_fn=ds.collate_fn)
    assert len(dl) == 3

    for batch, exp_len in zip(dl, [3, 3, 1]):
        assert batch.payload['cat'].size(0) == exp_len



import torch

from ptls.frames.coles.sampling_strategies.triplet_selector import TripletSelector
from ptls.frames.coles.metric import outer_pairwise_distance


class HardTripletSelector(TripletSelector):
    """
        Generate triplets with all positive pairs and the neg_count hardest negative example for each anchor
    """

    def __init__(self, neg_count=1):
        super(HardTripletSelector, self).__init__()
        self.neg_count = neg_count

    def get_triplets(self, embeddings, labels):
        n = labels.size(0)

        # construct matrix x, such as x_ij == 0 <==> labels[i] == labels[j]
        x = labels.expand(n, n) - labels.expand(n, n).t()

        positive_pairs = torch.triu((x == 0).int(), diagonal=1).nonzero(as_tuple=False)

        m = positive_pairs.size(0)

        anchor_embed = embeddings[positive_pairs[:, 0]].detach()
        anchor_labels = labels[positive_pairs[:, 0]]

        # pos_embed = embeddings[positive_pairs[:,0]].detach()

        # construct matrix x (size m x n), such as x_ij == 1 <==> anchor_labels[i] == labels[j]
        x = (labels.expand(m, n) == anchor_labels.expand(n, m).t())

        mat_distances = outer_pairwise_distance(anchor_embed, embeddings.detach())  # pairwise_distance anchors x all

        upper_bound = int((2 * n) ** 0.5) + 1
        mat_distances = ((upper_bound - mat_distances) * (x == 0).type(
            mat_distances.dtype))  # filter: get only negative pairs

        values, indices = mat_distances.topk(k=self.neg_count, dim=1, largest=True)

        triplets = torch.cat([
            positive_pairs.repeat(self.neg_count, 1),
            torch.cat(indices.unbind(dim=0)).view(-1, 1)
        ], dim=1)

        return triplets


import numpy as np
import torch

from ptls.data_load.iterable_processing_dataset import IterableProcessingDataset


class CategorySizeClip(IterableProcessingDataset):
    def __init__(self, category_max_size, replace_value='max'):
        """Deprecated. _smart_clip implemented in TrxEncoder

        Args:
            category_max_size: {field_name, max_size}
            replace_value: value for infrequent categories, int for specific value, 'max' for `category_max_size - 1`
        """
        super().__init__()

        self._category_max_size = category_max_size
        self._replace_value = replace_value

    def __iter__(self):
        for rec in self._src:
            features = rec[0] if type(rec) is tuple else rec

            # clip embeddings dictionary by max value
            for name, max_size in self._category_max_size.items():
                features[name] = self._smart_clip(features[name], max_size)
            yield rec

    def _smart_clip(self, values, max_size):
        if self._replace_value == 'max':
            return values.clip(0, max_size - 1)
        else:
            return torch.from_numpy(np.where((0 <= values) & (values < max_size), values, self._replace_value))



import torch
import numpy as np
from omegaconf import OmegaConf

from ptls.data_load.padded_batch import PaddedBatch
from ptls.nn.seq_encoder.rnn_seq_encoder_distribution_target import RnnSeqEncoderDistributionTarget
from ptls.nn.trx_encoder import TrxEncoder


def get_data():
    payload = {'amount': torch.arange(4*10).view(4, 10).float(),
               'event_time': torch.arange(4*10).view(4, 10).float(),
               'mcc_code': torch.arange(4*10).view(4, 10),
               'tr_type': torch.arange(4*10).view(4, 10)
              }
    return PaddedBatch(
                       payload=payload,
                       length=torch.tensor([4, 2, 6, 8])
                      )


def _test_shape():
    eps = 1e-5
    params = {
        'trx_encoder': {
            'norm_embeddings': False,
            'embeddings_noise': 0.003,
            'embeddings': {
                'mcc_code': {
                    'in': 200,
                    'out': 48
                },
                'tr_type': {
                    'in': 100,
                    'out': 24
                }
            },
            'numeric_values': {
                'amount': 'identity'
            },
        },
        'rnn': {
            'hidden_size': 48,
            'type': 'gru',
            'bidir': False,
            'trainable_starter': 'static'
        },
        'head_layers': {
            'CombinedTargetHeadFromRnn': {
                'in_size': 48,
                'num_distr_classes': 6,
                'pos': True,
                'neg': True,
                'use_gates': True,
                'pass_samples': True
            }
        }
    }
    params = OmegaConf.create(params)

    model = RnnSeqEncoderDistributionTarget(
        trx_encoder=TrxEncoder(**params['trx_encoder']),
        **params['rnn'],
        head_layers=params['head_layers'],
    )

    x = get_data()

    out = model(x)
    assert isinstance(out, tuple) and len(out) == 3
    assert isinstance(out[0], torch.Tensor) and out[0].shape == torch.Size([4, 48])
    assert (out[1] - np.array([-16.118095, -16.118095, -16.118095, -16.118095]) < np.zeros((1, 4)) + eps).all()
    assert (out[2] - np.array([3.302955, 11.313237, 25.456194, 37.45834])< np.zeros((1, 4)) + eps).all()



#!/usr/bin/env python

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()
    
setuptools.setup(
    name='pytorch-lifestream',
    version='0.5.2',
    author='',
    author_email='',
    description='Lifestream data analysis with PyTorch',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=[
        'pytorch-lightning==1.6.*',
        'pyarrow>=7.0',
        'transformers==4.*',
        'hydra-core>=1.1.2'
    ],
)



class TripletSelector:
    """
    Implementation should return indices of anchors, positive and negative samples
    return np array of shape [N_triplets x 3]
    """

    def __init__(self):
        pass

    def get_triplets(self, embeddings, labels):
        raise NotImplementedError


import pytorch_lightning as pl
import torch
from torch import nn
import warnings
from torchmetrics import MeanMetric
from typing import Tuple, Dict, List, Union

from ptls.nn.seq_encoder.abs_seq_encoder import AbsSeqEncoder
from ptls.nn import PBL2Norm
from ptls.data_load.padded_batch import PaddedBatch
from ptls.custom_layers import StatPooling, GEGLU
from ptls.nn.seq_step import LastStepEncoder

class Head(nn.Module):   
    def __init__(self, input_size, n_classes, hidden_size=64, drop_p=0.1):
        super().__init__()
        self.head = nn.Sequential(
            nn.Linear(input_size, hidden_size, bias=True),
            nn.GELU(),
            nn.Dropout(drop_p),
            nn.Linear(hidden_size, n_classes)
        )
    def forward(self, x):
        x = self.head(x)
        return x

class GptPretrainModule(pl.LightningModule):
    """GPT2 Language model

    Original sequence are encoded by `TrxEncoder`.
    Model `seq_encoder` predicts embedding of next transaction.
    Heads are used to predict each feature class of future transaction.

    Parameters
    ----------
    trx_encoder:
        Module for transform dict with feature sequences to sequence of transaction representations
    seq_encoder:
        Module for sequence processing. Generally this is transformer based encoder. Rnn is also possible
        Should works without sequence reduce
    head_hidden_size:
        Hidden size of heads for feature prediction
    seed_seq_len:
         Size of starting sequence without loss 
    total_steps:
        total_steps expected in OneCycle lr scheduler
    max_lr:
        max_lr of OneCycle lr scheduler
    weight_decay:
        weight_decay of Adam optimizer
    pct_start:
        % of total_steps when lr increase
    norm_predict:
        use l2 norm for transformer output or not
    inference_pooling_strategy:
        'out' - `seq_encoder` forward (`is_reduce_requence=True`) (B, H)
        'out_stat' - min, max, mean, std statistics pooled from `seq_encoder` layer (B, H) -> (B, 4H)
        'trx_stat' - min, max, mean, std statistics pooled from `trx_encoder` layer (B, H) -> (B, 4H)
        'trx_stat_out' - min, max, mean, std statistics pooled from `trx_encoder` layer + 'out' from `seq_encoder` (B, H) -> (B, 5H)
    """

    def __init__(self,
                 trx_encoder: torch.nn.Module,
                 seq_encoder: AbsSeqEncoder,
                 head_hidden_size: int = 64,
                 total_steps: int = 64000,
                 seed_seq_len: int = 16,
                 max_lr: float = 0.00005,
                 weight_decay: float = 0.0,
                 pct_start: float = 0.1,
                 norm_predict: bool = False,
                 inference_pooling_strategy: str = 'out_stat'
                 ):

        super().__init__()
        self.save_hyperparameters(ignore=['trx_encoder', 'seq_encoder'])

        self.trx_encoder = trx_encoder
        assert not self.trx_encoder.numeric_values, '`numeric_values` parameter of `trx_encoder` should be == {}. Discretize all numerical features into categorical to use Tabformer model!'
        assert self.trx_encoder.embeddings, '`embeddings` parameter for `trx_encoder` should contain at least 1 feature!'

        self._seq_encoder = seq_encoder
        self._seq_encoder.is_reduce_sequence = False

        self.head = nn.ModuleDict()
        for col_name, noisy_emb in self.trx_encoder.embeddings.items():
            self.head[col_name] = Head(input_size=self._seq_encoder.embedding_size, hidden_size=head_hidden_size, n_classes=noisy_emb.num_embeddings)

        if self.hparams.norm_predict:
            self.fn_norm_predict = PBL2Norm()

        self.loss = nn.CrossEntropyLoss(ignore_index=0)

        self.train_gpt_loss = MeanMetric()
        self.valid_gpt_loss = MeanMetric()

    def forward(self, batch: PaddedBatch):
        z_trx = self.trx_encoder(batch) 
        out = self._seq_encoder(z_trx)
        if self.hparams.norm_predict:
            out = self.fn_norm_predict(out)
        return out

    def loss_gpt(self, predictions, labels, is_train_step):
        loss = 0
        for col_name, head in self.head.items():
            y_pred = head(predictions[:, self.hparams.seed_seq_len:-1, :])
            y_pred = y_pred.view(-1, y_pred.size(-1))

            y_true = labels[col_name][:, self.hparams.seed_seq_len+1:]
            y_true = torch.flatten(y_true.long())

            loss += self.loss(y_pred, y_true)
        return loss

    def training_step(self, batch, batch_idx):
        out = self.forward(batch)  # PB: B, T, H
        out = out.payload if isinstance(out, PaddedBatch) else out
        labels = batch.payload

        loss_gpt = self.loss_gpt(out, labels, is_train_step=True)
        self.train_gpt_loss(loss_gpt)
        self.log(f'gpt/loss', loss_gpt, sync_dist=True)
        return loss_gpt

    def validation_step(self, batch, batch_idx):
        out = self.forward(batch)  # PB: B, T, H
        out = out.payload if isinstance(out, PaddedBatch) else out
        labels = batch.payload

        loss_gpt = self.loss_gpt(out, labels, is_train_step=False)
        self.valid_gpt_loss(loss_gpt)

    def training_epoch_end(self, _):
        self.log(f'gpt/train_gpt_loss', self.train_gpt_loss, prog_bar=False, sync_dist=True, rank_zero_only=True)
        # self.train_gpt_loss reset not required here

    def validation_epoch_end(self, _):
        self.log(f'gpt/valid_gpt_loss', self.valid_gpt_loss, prog_bar=True, sync_dist=True, rank_zero_only=True)
        # self.valid_gpt_loss reset not required here

    def configure_optimizers(self):
        optim = torch.optim.NAdam(self.parameters(),
                                 lr=self.hparams.max_lr,
                                 weight_decay=self.hparams.weight_decay,
                                 )
        scheduler = torch.optim.lr_scheduler.OneCycleLR(
            optimizer=optim,
            max_lr=self.hparams.max_lr,
            total_steps=self.hparams.total_steps,
            pct_start=self.hparams.pct_start,
            anneal_strategy='cos',
            cycle_momentum=False,
            div_factor=25.0,
            final_div_factor=10000.0,
            three_phase=False,
        )
        scheduler = {'scheduler': scheduler, 'interval': 'step'}
        return [optim], [scheduler]
    
    @property
    def seq_encoder(self):
        return GPTInferenceModule(pretrained_model=self)

class GPTInferenceModule(torch.nn.Module):
    def __init__(self, pretrained_model):
        super().__init__()
        self.model = pretrained_model
        self.model.is_reduce_sequence = False

        self.stat_pooler = StatPooling()
        self.last_step = LastStepEncoder()

    def forward(self, batch):
        z_trx = self.model.trx_encoder(batch)
        out = self.model._seq_encoder(z_trx)
        out = out if isinstance(out, PaddedBatch) else PaddedBatch(out, batch.seq_lens)
        if self.model.hparams.inference_pooling_strategy=='trx_stat_out':
            stats = self.stat_pooler(z_trx)
            out = self.last_step(out)
            out = torch.cat([stats, out], dim=1)
        elif self.model.hparams.inference_pooling_strategy=='trx_stat':
            out = self.stat_pooler(z_trx)
        elif self.model.hparams.inference_pooling_strategy=='out_stat':
            out = self.stat_pooler(out)
        elif self.model.hparams.inference_pooling_strategy=='out':
            out = self.last_step(out)
        else:
            raise
        if self.model.hparams.norm_predict:
            out = out / (out.pow(2).sum(dim=-1, keepdim=True) + 1e-9).pow(0.5)
        return out


import numpy as np
import torch
import logging

from ptls.data_load import IterableProcessingDataset

logger = logging.getLogger(__name__)


class SeqLenFilter(IterableProcessingDataset):
    def __init__(self, min_seq_len=None, max_seq_len=None, seq_len_col=None, sequence_col=None):
        """

        Args:
            min_seq_len: if set than drop sequences shorter than `min_seq_len`
            max_seq_len: if set than drop sequences longer than `max_seq_len`
            seq_len_col: field where sequence length stored, if None, `target_col` used
            sequence_col: field for sequence length detection, if None, any iterable field will be used
        """
        super().__init__()

        self._min_seq_len = min_seq_len
        self._max_seq_len = max_seq_len
        self._sequence_col = sequence_col
        self._seq_len_col = seq_len_col

    def __iter__(self):
        for rec in self._src:
            features = rec[0] if type(rec) is tuple else rec
            seq_len = self.get_len(features)
            if self._min_seq_len is not None and seq_len < self._min_seq_len:
                continue
            if self._max_seq_len is not None and seq_len > self._max_seq_len:
                continue
            yield rec

    def get_sequence_col(self, rec):
        if self._sequence_col is None:
            arrays = [k for k, v in rec.items() if self.is_seq_feature(k, v)]
            if len(arrays) == 0:
                raise ValueError(f'Can not find field with sequence from record: {rec}')
            self._sequence_col = arrays[0]
        return self._sequence_col

    def get_len(self, rec):
        if self._seq_len_col is not None:
            return rec[self._seq_len_col]
        return len(rec[self.get_sequence_col(rec)])



from ptls.data_load.iterable_processing_dataset import IterableProcessingDataset


class TargetJoin(IterableProcessingDataset):
    """Deprecated. Only single dict allowed.
    Store target as scalar value in a feature dictionary

    """
    def __init__(self, id_col, target_values, func=int):
        """Extract value from `target_values` by id_col and mention it as `y`

        for x, * in seq:
            id = x[id_col]
            y = target_values[id]
            yield x, y

        Args:
            id_col: field where `id` is stored
            target_values: dict with target

        """
        super().__init__()

        self._id_col = id_col
        self._target_values = target_values
        self.func = func

    def __iter__(self):
        for rec in self._src:
            features = rec[0] if type(rec) is tuple else rec
            _id = features[self._id_col]
            y = self.func(self._target_values[_id])
            yield features, y



import logging

from functools import reduce
from operator import add
from torch.utils.data import Dataset

logger = logging.getLogger(__name__)


class ComplexTargetDataset(Dataset):
    def __init__(self, base_dataset, split_counts):
        self.base_dataset = base_dataset
        self.split_counts = split_counts

    def __len__(self):
        return len(self.base_dataset)

    def __getitem__(self, idx):
        row = self.base_dataset[idx]
        splitter_numbers = [[i]*split_count for i, split_count in enumerate(self.split_counts)]
        splitter_numbers = reduce(add, splitter_numbers)
        data = [(x, (n, idx)) for x, n in zip(row, splitter_numbers)]
        return data



import torch

from ptls.data_load.padded_batch import PaddedBatch
from ptls.nn.seq_encoder import TransformerEncoder
from ptls.nn.seq_encoder.transformer_encoder import PositionalEncoding


def test_positional_encoding_shape():
    pe = PositionalEncoding(256, max_len=4000)
    assert pe.pe.size() == (1, 4000, 256)


def test_positional_encoding_forward_train():
    pe = PositionalEncoding(256, max_len=4000)
    pe.train()
    x = torch.randn(10, 128, 256)
    y = pe(x)
    assert y.size() == (10, 128, 256)


def test_positional_encoding_forward_eval():
    pe = PositionalEncoding(256, 4000)
    pe.eval()
    x = torch.randn(10, 128, 256)
    y = pe(x)
    assert y.size() == (10, 128, 256)


def test_transformer_encoder_example():
    model = TransformerEncoder(input_size=32)
    x = PaddedBatch(torch.randn(10, 128, 32), torch.randint(20, 128, (10,)))
    y = model(x)
    assert y.payload.size() == (10, 128, 32)

    model = TransformerEncoder(input_size=32, is_reduce_sequence=True)
    y = model(x)
    assert y.size() == (10, 32)


def test_transformer_embedding_size():
    model = TransformerEncoder(input_size=32)
    assert model.embedding_size == 32


def test_transformer_params():
    x = PaddedBatch(torch.randn(10, 128, 32), torch.randint(20, 128, (10,)))
    models = [
        TransformerEncoder(input_size=32, starter='zeros'),
        TransformerEncoder(input_size=32, shared_layers=True),
        TransformerEncoder(input_size=32, use_positional_encoding=False),
        TransformerEncoder(input_size=32, use_start_random_shift=False),
        TransformerEncoder(input_size=32, use_after_mask=True),
        TransformerEncoder(input_size=32, use_src_key_padding_mask=False),
        TransformerEncoder(input_size=32, use_norm_layer=False),
        TransformerEncoder(input_size=32, use_norm_layer=False, shared_layers=True),
    ]
    for model in models:
        y = model(x)
        assert y.payload.size() == (10, 128, 32)



import numpy as np
import torch
from ptls.data_load.iterable_processing_dataset import IterableProcessingDataset


class ToTorch(IterableProcessingDataset):
    """Filter that transforms each numpy.ndarray in sample dict to torch.Tensor
         
    Filter transformation:
    list({key1: int, key2: np.ndarray, ...}) -> list({key1: int, key2: torch.Tensor, ...})
    (in place of list there could be other iterable, for example generator)
    """
    def __init__(self):
        super().__init__()

    def __iter__(self):
        for rec in self._src:
            features = rec[0] if type(rec) is tuple else rec
            for k, v in features.items():
                if isinstance(v, np.ndarray) and v.dtype.kind in ('i', 'f'):
                    features[k] = torch.from_numpy(v)
            yield rec



from ptls.preprocessing.base import ColTransformer
from ptls.preprocessing.pandas.col_transformer import ColTransformerPandasMixin


class ColIdentityEncoder(ColTransformerPandasMixin, ColTransformer):
    """Dont change original column

    May rename or copy original columns.

    Parameters
    ----------
    col_name_original:
        Source column name
    col_name_target:
        Target column name. Transformed column will be placed here
        If `col_name_target is None` then original column will be replaced by transformed values.
    is_drop_original_col:
        When target and original columns are different manage original col deletion.

    """
    def transform(self, x):
        x = self.attach_column(x, x[self.col_name_original].rename(self.col_name_target))

        x = super().transform(x)
        return x



import torch

from ptls.data_load.utils import collate_feature_dict


class SeqToTargetDataset(torch.utils.data.Dataset):
    def __init__(self,
                 data,
                 target_col_name,
                 target_dtype=None,
                 *args, **kwargs,
                 ):
        super().__init__(*args, **kwargs)

        self.data = data
        self.target_col_name = target_col_name
        if type(target_dtype) is str:
            self.target_dtype = getattr(torch, target_dtype)
        else:
            self.target_dtype = target_dtype

    def __len__(self):
        return len(self.data)

    def __getitem__(self, item):
        feature_arrays = self.data[item]
        return feature_arrays

    def __iter__(self):
        for feature_arrays in self.data:
            yield feature_arrays


    def collate_fn(self, padded_batch):
        padded_batch = collate_feature_dict(padded_batch)
        target = padded_batch.payload[self.target_col_name]
        del padded_batch.payload[self.target_col_name]
        if self.target_dtype is not None:
            target = target.to(dtype=self.target_dtype)
        return padded_batch, target


class SeqToTargetIterableDataset(SeqToTargetDataset, torch.utils.data.IterableDataset):
    pass



import numpy as np
import torch

from ptls.data_load.iterable_processing_dataset import IterableProcessingDataset


class FeatureFilter(IterableProcessingDataset):
    def __init__(self, keep_feature_names=None, drop_feature_names=None, drop_non_iterable=True):
        """

        Args:
            keep_feature_names: feature name for keep
        """
        super().__init__()

        if type(keep_feature_names) is str:
            keep_feature_names = [keep_feature_names]
        if type(drop_feature_names) is str:
            drop_feature_names = [drop_feature_names]

        self._keep_feature_names = set(keep_feature_names) if keep_feature_names is not None else None
        self._drop_feature_names = set(drop_feature_names) if drop_feature_names is not None else None

        self._drop_non_iterable = drop_non_iterable

    def process(self, features):
        if self._drop_feature_names is not None:
            features = {k: v for k, v in features.items()
                        if k not in self._drop_feature_names or self.is_keep(k)}
        if self._drop_non_iterable:
            features = {k: v for k, v in features.items()
                        if self.is_seq_feature(k, v) or self.is_keep(k)}
        return features

    def is_keep(self, k):
        if self._keep_feature_names is None:
            return False
        return k in self._keep_feature_names



import torch
from torch import nn as nn


class L2NormEncoder(nn.Module):
    def __init__(self, eps=1e-9):
        super().__init__()
        self.eps = eps

    def forward(self, x: torch.Tensor):
        return x / (x.pow(2).sum(dim=-1, keepdim=True) + self.eps).pow(0.5)



import numpy as np
import pytest
import torch

from ptls.data_load.padded_batch import PaddedBatch


def get_pb():
    return PaddedBatch(
        payload={
            'bin': torch.IntTensor([0, 1]),
            'target_bin': torch.IntTensor([2, 3]),
            'pp': torch.FloatTensor([0.1, 0.2]),
            'user_id': np.array(['a', 'b']),
            'lists': np.array([[5, 6], [7, 8]]),
            'mcc': torch.tensor([
                [1, 2, 0, 0],
                [3, 4, 5, 6],
            ]),
            'event_time': torch.tensor([
                [1, 2, 0, 0],
                [1, 2, 3, 4],
            ]),
            'target_array': torch.tensor([
                [1, 2, 2, 5],
                [1, 2, 3, 4],
            ]),
        },
        length=torch.IntTensor([2, 4])
    )

def get_pb_with_text_event_time():
    return PaddedBatch(
        payload={
            'bin': torch.IntTensor([0, 1]),
            'target_bin': torch.IntTensor([2, 3]),
            'pp': torch.FloatTensor([0.1, 0.2]),
            'user_id': np.array(['a', 'b']),
            'lists': np.array([[5, 6], [7, 8]]),
            'mcc': torch.tensor([
                [1, 2, 0, 0],
                [3, 4, 5, 6],
            ]),
            'event_time': np.array([
                ['0 10:00:05', '0 11:00:00', '2 11:12:10', '2 12:00:00'],
                ['1 10:00:05', '3 11:00:00', '4 11:12:10', '4 12:00:00'],
            ]),
            'target_array': torch.tensor([
                [1, 2, 2, 5],
                [1, 2, 3, 4],
            ]),
        },
        length=torch.IntTensor([2, 4])
    )

def test_padded_batch_example():
    data = PaddedBatch(
        payload=torch.tensor([
            [1, 2, 0, 0],
            [3, 4, 5, 6],
            [7, 8, 9, 0],
        ]),
        length=torch.tensor([2, 4, 3]),
    )

    # check shape
    torch.testing.assert_close(data.payload.size(), (3, 4))

    # get first transaction
    torch.testing.assert_close(data.payload[:, 0], torch.tensor([1, 3, 7]))

    # get last transaction
    torch.testing.assert_close(data.payload[torch.arange(3), data.seq_lens - 1], torch.tensor([2, 6, 9]))

    # get all transaction flatten
    torch.testing.assert_close(data.payload[data.seq_len_mask.bool()], torch.tensor([1, 2, 3, 4, 5, 6, 7, 8, 9]))


def test_padded_batch_seq_feature_shape():
    x = get_pb()
    B, T = x.seq_feature_shape
    assert B, T == (2, 4)


def test_padded_batch_to():
    x = get_pb()
    y = x.to('cpu')
    assert len(y) == 2


def test_padded_batch_mask_tensor_trx_embedding():
    data = PaddedBatch(torch.randn(4, 5, 3), torch.tensor([2, 5, 1, 3]))
    out = data.seq_len_mask
    exp = torch.tensor([
        [1, 1, 0, 0, 0],
        [1, 1, 1, 1, 1],
        [1, 0, 0, 0, 0],
        [1, 1, 1, 0, 0],
    ]).long()
    torch.testing.assert_close(out, exp)


def test_padded_batch_mask_tensor_numerical():
    data = PaddedBatch(torch.randn(4, 5), torch.tensor([2, 5, 3, 1]))
    out = data.seq_len_mask
    exp = torch.tensor([
        [1, 1, 0, 0, 0],
        [1, 1, 1, 1, 1],
        [1, 1, 1, 0, 0],
        [1, 0, 0, 0, 0],
    ]).long()
    torch.testing.assert_close(out, exp)


def test_padded_batch_mask_dict():
    data = PaddedBatch({'col1': torch.randn(4, 5), 'col2': torch.randn(4, 5)}, torch.tensor([4, 5, 1, 3]))
    out = data.seq_len_mask
    exp = torch.tensor([
        [1, 1, 1, 1, 0],
        [1, 1, 1, 1, 1],
        [1, 0, 0, 0, 0],
        [1, 1, 1, 0, 0]
    ]).long()
    torch.testing.assert_close(out, exp)


def test_padded_batch_mask_dict_no_sequences():
    with pytest.raises(StopIteration):
        data = PaddedBatch({'target_col1': torch.randn(4, 5), 'bin': torch.IntTensor([2, 3, 1, 2])}, torch.tensor([4, 5, 1, 3]))
        _ = data.seq_len_mask

# is seq feature
def test_padded_batch_is_seq_feature():
    x = get_pb()

    for col, is_seq in [
        ('bin', False),
        ('target_bin', False),
        ('pp', False),
        ('user_id', False),
        ('lists', False),
        ('mcc', True),
        ('event_time', True),
        ('target_array', False),
    ]:
        assert is_seq == PaddedBatch.is_seq_feature(col, x.payload[col]), col

# is seq feature
def test_padded_batch_is_seq_feature_with_text_event_time():
    x_text = get_pb_with_text_event_time()

    for col, is_seq in [
        ('bin', False),
        ('target_bin', False),
        ('pp', False),
        ('user_id', False),
        ('lists', False),
        ('mcc', True),
        ('event_time', True),
        ('target_array', False),
    ]:
        assert is_seq == PaddedBatch.is_seq_feature(col, x_text.payload[col]), col


def test_padded_batch_drop_seq_features():
    x = get_pb()
    y = x.drop_seq_features()

    for col, is_seq in [
        ('bin', False),
        ('target_bin', False),
        ('pp', False),
        ('user_id', False),
        ('lists', False),
        ('mcc', True),
        ('event_time', True),
        ('target_array', False),
    ]:
        assert is_seq != (col in y)


def test_padded_batch_keep_seq_features():
    x = get_pb()
    y = x.keep_seq_features()

    for col, is_seq in [
        ('bin', False),
        ('target_bin', False),
        ('pp', False),
        ('user_id', False),
        ('lists', False),
        ('mcc', True),
        ('event_time', True),
        ('target_array', False),
    ]:
        assert is_seq == (col in y.payload)



import torch
from ptls.data_load.padded_batch import PaddedBatch
from ptls.nn import (
    PBLinear, PBL2Norm, PBLayerNorm, PBReLU,
)


def test_pb_doc_example():
    x = PaddedBatch(torch.randn(4, 12, 8), torch.LongTensor([3, 12, 8]))
    model = torch.nn.Sequential(
        PBLinear(8, 5),
        PBReLU(),
        PBLinear(5, 10),
    )
    y = model(x)
    assert y.payload.size() == (4, 12, 10)


def test_pb_linear_args():
    x = PaddedBatch(torch.randn(4, 12, 8), torch.LongTensor([3, 12, 8]))
    model = PBLinear(8, 5)
    y = model(x)
    assert y.payload.size() == (4, 12, 5)
    help(PBLinear)


def test_pb_linear_kwargs():
    x = PaddedBatch(torch.randn(4, 12, 8), torch.LongTensor([3, 12, 8]))
    model = PBLinear(in_features=8, out_features=5)
    y = model(x)
    assert y.payload.size() == (4, 12, 5)


def test_pb_layer_norm():
    x = PaddedBatch(torch.randn(4, 12, 8), torch.LongTensor([3, 12, 8]))
    model = PBLayerNorm(8)
    y = model(x)
    assert y.payload.size() == (4, 12, 8)


def test_pb_l2_norm():
    x = PaddedBatch(torch.randn(4, 12, 8), torch.LongTensor([3, 12, 8]))
    model = PBL2Norm()
    y = model(x)
    assert y.payload.size() == (4, 12, 8)
    torch.testing.assert_allclose(torch.ones(4, 12), y.payload.pow(2).sum(dim=2))


def test_pb_relu():
    x = PaddedBatch(torch.randn(4, 12, 8), torch.LongTensor([3, 12, 8]))
    model = PBReLU()
    y = model(x)
    assert y.payload.size() == (4, 12, 8)



import torch
from pyhocon import ConfigFactory

from ptls.nn.seq_encoder import AggFeatureSeqEncoder
from ptls.data_load.padded_batch import PaddedBatch


def get_data():
    return PaddedBatch(
        payload={
            'cat1': torch.tensor([
                [1, 3, 5, 0, 0, 0],
                [2, 5, 3, 1, 1, 0],
                [4, 0, 0, 0, 0, 0],
            ]),
            'num1': torch.tensor([
                [5, 6, 3, 0, 0, 0],
                [4, 0, 1, 3, 2, 0],
                [7, 0, 0, 0, 0, 0],
            ]).float(),
        },
        length=torch.tensor([3, 5, 1]),
    )


def get_conf():
    params = {
        'numeric_values': {
            'num1': {},
        },
        'embeddings': {
            'cat1': {'in': 8}
        },
        'was_logified': False,
        'log_scale_factor': 1.0,
        'is_used_count': False,
        'is_used_mean': True,
        'is_used_std': True,
        "use_topk_cnt": 3,
    }
    return ConfigFactory.from_dict(params)


def test_output_size():
    embedding_size = AggFeatureSeqEncoder(**get_conf()).embedding_size
    assert embedding_size == 24


def test_model():
    model = AggFeatureSeqEncoder(**get_conf())
    out = model(get_data())
    assert out.size() == (3, 24)



from functools import reduce
from operator import iadd
from typing import List

import numpy as np
import torch

from ptls.data_load.feature_dict import FeatureDict
from ptls.data_load.utils import collate_feature_dict
from ptls.frames.coles.split_strategy import AbsSplit


class ColesSupervisedDataset(FeatureDict, torch.utils.data.Dataset):
    """Dataset for ptls.frames.coles.ColesSupervisedModule

    Parameters
    ----------
    data:
        source data with feature dicts
    splitter:
        object from from `ptls.frames.coles.split_strategy`.
        Used to split original sequence into subsequences which are samples from one client.
    cols_classes:
        column names with class labels for auxiliary supervised loss calculation
    col_time:
        column name with event_time
    """
    def __init__(self,
                 data,
                 splitter: AbsSplit,
                 cols_classes: List[str],
                 col_time='event_time',
                 *args, **kwargs):
        super().__init__(*args, **kwargs)  # required for mixin class

        self.data = data
        self.splitter = splitter
        self.col_time = col_time
        self.cols_classes = cols_classes

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        feature_arrays = self.data[idx]
        return self.get_splits(feature_arrays), self.get_classes(feature_arrays)

    def __iter__(self):
        for feature_arrays in self.data:
            yield self.get_splits(feature_arrays), self.get_classes(feature_arrays)

    def get_splits(self, feature_arrays):
        local_date = feature_arrays[self.col_time]
        indexes = self.splitter.split(local_date)
        return [{k: v[ix] for k, v in feature_arrays.items() if self.is_seq_feature(k, v)} for ix in indexes]

    def get_classes(self, feature_arrays):
        res = [feature_arrays.get(col, -1) for col in self.cols_classes]
        res = [r if ~np.isnan(r) else -1 for r in res]
        return res

    @staticmethod
    def collate_fn(batch):
        class_labels = [i for i, (seq, labels) in enumerate(batch) for _ in seq]
        seq_samples = [seq for seq, labels in batch]
        target_labels = [labels for seq, labels in batch for _ in seq]
        batch = reduce(iadd, seq_samples)
        padded_batch = collate_feature_dict(batch)
        return padded_batch, torch.LongTensor(class_labels), torch.LongTensor(target_labels)


class ColesSupervisedIterableDataset(ColesSupervisedDataset, torch.utils.data.IterableDataset):
    pass



from typing import Dict

import torch as torch
import warnings

from ptls.nn.trx_encoder.trx_encoder import TrxEncoder


class TrxEncoderOhe(TrxEncoder):
    """Network layer which makes representation for single transactions
    Based on `ptls.nn.trx_encoder.TrxEncoder` with embedding one-hot initialization.

     Input is `PaddedBatch` with ptls-format dictionary, with feature arrays of shape (B, T)
     Output is `PaddedBatch` with transaction embeddings of shape (B, T, H)
     where:
        B - batch size, sequence count in batch
        T - sequence length
        H - hidden size, representation dimension

    `ptls.nn.trx_encoder.noisy_embedding.NoisyEmbedding` implementation are used for categorical features.

    Parameters
        embeddings:
            dict with categorical feature names.
            Values must be like this `{'in': dictionary_size}`
            These features will be encoded with lookup embedding table of shape (dictionary_size, embedding_size)
        is_learnable:
            correspond `requires_grad` attribute of embeddings parameters.
            Linear projection always learnable (if exists)
        numeric_values:
            dict with numerical feature names.
            Values must be a string with scaler_name.
            Possible values are: 'identity', 'sigmoid', 'log', 'year'.
            These features will be scaled with selected scaler.
            Values can be `ptls.nn.trx_encoder.scalers.BaseScaler` implementatoin

            One field can have many scalers. In this case key become alias and col name should be in scaler.
            Check `TrxEncoderBase.numeric_values` for more details

        use_batch_norm:
            True - All numerical values will be normalized after scaling
            False - No normalizing for numerical values
        use_batch_norm_with_lens:
            True - Respect seq_lens during batch_norm. Padding zeroes will be ignored
            False - Batch norm ower all time axis. Padding zeroes will included.

        linear_projection_size:
            Linear layer at the end will be added for non-zero value

        out_of_index:
            How to process a categorical indexes which are greater than dictionary size.
            'clip' - values will be collapsed to maximum index. This works well for frequency encoded categories.
                We join infrequent categories to one.
            'assert' - raise an error of invalid index appear.

    Examples:
        >>> B, T = 5, 20
        >>> trx_encoder = TrxEncoderOhe(
        >>>     embeddings={'mcc_code': {'in': 100}},
        >>>     numeric_values={'amount': 'log'},
        >>> )
        >>> x = PaddedBatch(
        >>>     payload={
        >>>         'mcc_code': torch.randint(0, 99, (B, T)),
        >>>         'amount': torch.randn(B, T),
        >>>     },
        >>>     length=torch.randint(10, 20, (B,)),
        >>> )
        >>> z = trx_encoder(x)
        >>> assert z.payload.shape == (5, 20, 100)  # B, T, H
    """

    def __init__(self,
                 embeddings: Dict[str, Dict] = None,
                 is_learnable: bool = False,
                 numeric_values: Dict[str, Dict] = None,
                 use_batch_norm=True,
                 use_batch_norm_with_lens=False,
                 linear_projection_size=0,
                 out_of_index: str = 'clip',
                 ):

        for k, v in embeddings.items():
            new_out = v['in'] - 1
            if 'out' in v:
                if v['out'] != new_out:
                    warnings.warn(f'Embedding out correction. Embeddings out for "{k}" should be equal `in - 1`, '
                                  f'found {v["out"]}, should be {new_out}. '
                                  f'Set correct value for `embeddings.{k}.out` or drop it to use automatic output set',
                                  UserWarning)
            v['out'] = new_out

        super().__init__(
            embeddings=embeddings,
            numeric_values=numeric_values,
            out_of_index=out_of_index,
            use_batch_norm=use_batch_norm,
            use_batch_norm_with_lens=use_batch_norm_with_lens,
            linear_projection_size=linear_projection_size,
            embeddings_noise=0,
            emb_dropout=0,
            spatial_dropout=False,
            orthogonal_init=False,
        )

        for n, p in self.named_parameters():
            if n.startswith('embeddings.') and n.endswith('.weight'):
                torch.nn.init.eye_(p.data[1:])
                p.requires_grad = is_learnable



from torch import nn as nn
from torch.nn import functional as F
from ptls.frames.coles.sampling_strategies import PairwiseMatrixSelector


class SoftmaxPairwiseLoss(nn.Module):
    """
    Softmax Pairwise loss.
    """

    def __init__(self, pair_selector=None, temperature=0.05, eps=1e-6):
        super(SoftmaxPairwiseLoss, self).__init__()
        self.pair_selector = pair_selector
        if pair_selector is None:
            self.pair_selector = PairwiseMatrixSelector()
        self.temperature = temperature
        self.eps = eps

    def forward(self, embeddings, classes):
        pair_matrix = self.pair_selector.get_pair_matrix(embeddings, classes)
        similarities = F.cosine_similarity(pair_matrix[:, :, 0, :], pair_matrix[:, :, 1, :], dim=-1, eps=self.eps)
        similarities /= self.temperature
        log_matrix = (-1)*F.log_softmax(similarities)
        loss = log_matrix/(len(similarities))
        return loss[:, :1].sum()



import torch
import torchmetrics

from ptls.loss import BCELoss
from ptls.frames.abs_module import ABSModule


class SentencePairsHead(torch.nn.Module):
    def __init__(self, base_model, embeds_dim, hidden_size, drop_p):
        super().__init__()
        self.base_model = base_model
        # TODO: check batch_norm position
        self.head = torch.nn.Sequential(
            torch.nn.Linear(embeds_dim * 2, hidden_size, bias=True),
            torch.nn.BatchNorm1d(hidden_size),
            torch.nn.ReLU(),
            torch.nn.Dropout(drop_p),
            # torch.nn.BatchNorm1d(config['hidden_size']),
            torch.nn.Linear(hidden_size, 1),
            torch.nn.Sigmoid()
        )

    def forward(self, x):
        left, right = x
        x = torch.cat([self.base_model(left), self.base_model(right)], dim=1)
        return self.head(x).squeeze(-1)


class SopNspModule(ABSModule):
    """Next Sequence Prediction (NSP) from [BERT](https://arxiv.org/abs/1810.04805)
    Sequences Order Prediction (SOP) from [ALBERT](https://arxiv.org/abs/1909.11942)

    Task depends on target in dataloader.

    Original sequence are split into two parts. Next `seq_encoder` makes embeddings.
    For NSP target `head` predict is a right part a correct continuation of left part.
    Correct means sampled from tha same original sequence in right order.
    Incorrect mean sampled from different original sequence.
    For SOP target `head` predict is a left and right part are in the correct order.
    Left an rights part are sampled from one sequence.

    """

    def __init__(self, validation_metric=None,
                       seq_encoder=None,
                       hidden_size=None, drop_p=None,
                       loss=None,
                       optimizer_partial=None,
                       lr_scheduler_partial=None):

        if loss is None:
            loss = BCELoss()
        if validation_metric is None:
            validation_metric = torchmetrics.AUROC(num_classes=2)

        super().__init__(validation_metric,
                         seq_encoder,
                         loss,
                         optimizer_partial,
                         lr_scheduler_partial)

        self._head = SentencePairsHead(self.seq_encoder, self.seq_encoder.embedding_size, hidden_size, drop_p)

    @property
    def metric_name(self):
        return 'valid_auroc'

    @property
    def is_requires_reduced_sequence(self):
        return True

    def shared_step(self, x, y):
        y_h = self._head(x)
        return y_h, y



import pandas as pd

from ptls.preprocessing.base import ColTransformer
from ptls.preprocessing.pandas.col_transformer import ColTransformerPandasMixin


def dt_to_timestamp(x: pd.Series):
    return pd.to_datetime(x).astype('datetime64[s]').astype('int64') // 1000000000


def timestamp_to_dt(x: pd.Series):
    return pd.to_datetime(x, unit='s')


class DatetimeToTimestamp(ColTransformerPandasMixin, ColTransformer):
    """Converts datetime column to unix timestamp

    Parameters
    ----------
    col_name_original:
        Source column name
    col_name_target:
        Is always `event_time`. This is predefined column name.
    is_drop_original_col:
        When target and original columns are different manage original col deletion.

    """
    def __init__(self,
                 col_name_original: str = 'event_time',
                 is_drop_original_col: bool = True,
                 ):
        super().__init__(
            col_name_original=col_name_original,
            col_name_target='event_time',
            is_drop_original_col=is_drop_original_col,
        )

    def transform(self, x: pd.DataFrame):
        x = self.attach_column(x, dt_to_timestamp(x[self.col_name_original]).rename(self.col_name_target))
        x = super().transform(x)
        return x
